#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : voxel_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 07.01.2020


import  pickle
import  time
import  textwrap
from    concurrent.futures      import  ThreadPoolExecutor

import  numpy                   as      np

import  plotly.graph_objects    as      go
import  matplotlib
import  matplotlib.pyplot       as      plt

from    tqdm                    import  tqdm

import  pept
from    pept.utilities.traverse import  traverse3d




class Voxels(np.ndarray):
    '''A class that manages a single 3D voxel space, including tools for voxel
    traversal of lines, manipulation and visualisation.

    This class can be instantiated in a couple of ways:

    1. The constructor receives a pre-defined voxel space (i.e. a 3D numpy
       array), along with the space boundaries `xlim`, `ylim` and `zlim`.

    2. The `from_lines` method receives a sample of 3D lines (i.e. a 2D numpy
       array), each defined by two points, creating a voxel space and
       traversing / voxellising the lines.

    3. The `empty` method creates a voxel space filled with zeros.

    This subclasses the `numpy.ndarray` class, so any `Voxels` object acts
    exactly like a 3D numpy array. All numpy methods and operations are valid
    on `Voxels` (e.g. add 1 to all voxels with `voxels += 1`).

    It is possible to add multiple samples of lines to the same voxel space
    using the `add_lines` method.

    If you want to voxellise multiple samples of lines, see the `VoxelData`
    class.

    Attributes
    ----------
    voxels: (M, N, P) numpy.ndarray
        The 3D numpy array containing the number of lines that pass through
        each voxel. They are stored as `float`s. This class assumes a uniform
        grid of voxels - that is, the voxel size in each dimension is constant,
        but can vary from one dimension to another. The number of voxels in
        each dimension is defined by `number_of_voxels`.

    number_of_voxels: 3-tuple
        A 3-tuple corresponding to the shape of `voxels`.

    voxel_size: (3,) numpy.ndarray
        The lengths of a voxel in the x-, y- and z-dimensions, respectively.

    xlim: (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        x-dimension, formatted as [x_min, x_max].

    ylim: (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        y-dimension, formatted as [y_min, y_max].

    zlim: (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        z-dimension, formatted as [z_min, z_max].

    voxel_grids: list[numpy.ndarray]
        A list containing the voxel gridlines in the x-, y-, and z-dimensions.
        Each dimension's gridlines are stored as a numpy of the voxel
        delimitations, such that it has length (M + 1), where M is the number
        of voxels in given dimension.

    Methods
    -------
    save(filepath)
        Save a `Voxels` instance as a binary `pickle` object.

    load(filepath)
        Load a saved / pickled `Voxels` object from `filepath`.

    from_lines(lines, number_of_voxels, xlim = None, ylim = None, zlim = None,\
               verbose = True)
        Create a voxel space and traverse / voxellise a given sample of
        `lines`.

    empty(number_of_voxels, xlim, ylim, zlim, verbose = False)
        Create an empty voxel space for the 3D cube bounded by `xlim`,
        `ylim` and `zlim`.

    get_cutoff(p1, p2)
        Return a numpy array containing the minimum and maximum value found
        across the two input arrays.

    add_lines(lines, verbose = False)
        Voxellise a sample of lines, adding 1 to each voxel traversed, for
        each line in the sample.

    cube_trace(index, color = None, opacity = 0.4, colorbar = True,\
               colorscale = "magma")
        Get the Plotly `Mesh3d` trace for a single voxel at `index`.

    cubes_traces(condition = lambda voxels: voxels > 0, color = None,\
                 opacity = 0.4, colorbar = True, colorscale = "magma")
        Get a list of Plotly `Mesh3d` traces for all voxels selected by the
        `condition` filtering function.

    voxels_trace(condition = lambda voxel_data: voxel_data > 0, size = 4,\
                 color = None, opacity = 0.4, colorbar = True,\
                 colorscale = "Magma", colorbar_title = None)
        Create and return a trace for all the voxels in this class, with
        possible filtering.

    heatmap_trace(ix = None, iy = None, iz = None, width = 0,\
                  colorscale = "Magma", transpose = True)
        Create and return a Plotly `Heatmap` trace of a 2D slice through the
        voxels.

    Notes
    -----
    The traversed lines do not need to be fully bounded by the voxel space.
    Their intersection is automatically computed.

    The class saves `voxels` as a **contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the shape of the array after instantiating the class.

    Examples
    --------
    This class is most often instantiated from a sample of lines to voxellise:

    >>> import pept
    >>> import numpy as np

    >>> lines = np.arange(70).reshape(10, 7)

    >>> number_of_voxels = [3, 4, 5]
    >>> voxels = pept.Voxels.from_lines(lines, number_of_voxels)
    >>> Initialised Voxels class in 0.0006861686706542969 s.

    >>> print(voxels)
    >>> voxels:
    >>> [[[2. 1. 0. 0. 0.]
    >>>   [0. 2. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]]

    >>>  [[0. 0. 0. 0. 0.]
    >>>   [0. 1. 1. 0. 0.]
    >>>   [0. 0. 1. 1. 0.]
    >>>   [0. 0. 0. 0. 0.]]

    >>>  [[0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 2. 0.]
    >>>   [0. 0. 0. 1. 2.]]]

    >>> number_of_voxels =    (3, 4, 5)
    >>> voxel_size =          [22.  16.5 13.2]

    >>> xlim =                [ 1. 67.]
    >>> ylim =                [ 2. 68.]
    >>> zlim =                [ 3. 69.]

    >>> voxel_grids:
    >>> [array([ 1., 23., 45., 67.]),
    >>>  array([ 2. , 18.5, 35. , 51.5, 68. ]),
    >>>  array([ 3. , 16.2, 29.4, 42.6, 55.8, 69. ])]

    Note that it is important to define the `number_of_voxels`.

    See Also
    --------
    pept.VoxelData : Asynchronously manage multiple voxel spaces.
    pept.LineData : Encapsulate lines for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    def __new__(
        cls,
        voxels_array,
        xlim,
        ylim,
        zlim,
    ):
        '''`Voxels` class constructor.

        Parameters
        ----------
        voxels_array: 3D numpy.ndarray
            A 3D numpy array, corresponding to a pre-defined voxel space.

        xlim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            x-dimension, formatted as [x_min, x_max].

        ylim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            y-dimension, formatted as [y_min, y_max].

        zlim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            z-dimension, formatted as [z_min, z_max].

        Raises
        ------
        ValueError
            If `voxels_array` does not have exactly 3 dimensions or if
            `xlim`, `ylim` or `zlim` do not have exactly 2 values each.

        '''

        # Type-checking inputs
        voxels_array = np.asarray(
            voxels_array,
            order = "C",
            dtype = float
        )

        if voxels_array.ndim != 3:
            raise ValueError(textwrap.fill((
                "The input `voxels_array` must contain an array-like with "
                "exactly three dimensions (i.e. pre-made voxels array). "
                f"Received an array with {voxels_array.ndim} dimensions. "
                "Note: if you would like to create voxels from a sample of"
                "lines, use the `Voxels.from_lines` method. "
            )))

        xlim = np.asarray(xlim, dtype = float)

        if xlim.ndim != 1 or len(xlim) != 2:
            raise ValueError(textwrap.fill((
                "The input `xlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the voxel space in the x-dimension. "
                f"Received parameter with shape {xlim.shape}."
            )))

        ylim = np.asarray(ylim, dtype = float)

        if ylim.ndim != 1 or len(ylim) != 2:
            raise ValueError(textwrap.fill((
                "The input `ylim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the voxel space in the y-dimension. "
                f"Received parameter with shape {ylim.shape}."
            )))

        zlim = np.asarray(zlim, dtype = float)

        if zlim.ndim != 1 or len(zlim) != 2:
            raise ValueError(textwrap.fill((
                "The input `zlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the voxel space in the z-dimension. "
                f"Received parameter with shape {zlim.shape}."
            )))

        # Setting class attributes
        voxels = voxels_array.view(cls)
        voxels._number_of_voxels = voxels.shape

        voxels._xlim = xlim
        voxels._ylim = ylim
        voxels._zlim = zlim

        voxels._voxel_size = np.array([
            (voxels._xlim[1] - voxels._xlim[0]) / voxels._number_of_voxels[0],
            (voxels._ylim[1] - voxels._ylim[0]) / voxels._number_of_voxels[1],
            (voxels._zlim[1] - voxels._zlim[0]) / voxels._number_of_voxels[2],
        ])

        voxels._voxel_grids = tuple([
            np.linspace(lim[0], lim[1], voxels._number_of_voxels[i] + 1)
            for i, lim in enumerate((voxels._xlim, voxels._ylim, voxels._zlim))
        ])

        return voxels


    def __array_finalize__(self, voxels):
        if voxels is None:
            return

        self._number_of_voxels = getattr(voxels, "_number_of_voxels", None)
        self._voxel_size = getattr(voxels, "_voxel_size", None)

        self._xlim = getattr(voxels, "_xlim", None)
        self._ylim = getattr(voxels, "_ylim", None)
        self._zlim = getattr(voxels, "_zlim", None)

        self._voxel_grids = getattr(voxels, "_voxel_grids", None)


    def __reduce__(self):
        # __reduce__ and __setstate__ ensure correct pickling behaviour. See
        # https://stackoverflow.com/questions/26598109/preserve-custom-
        # attributes-when-pickling-subclass-of-numpy-array

        # Get the parent's __reduce__ tuple
        pickled_state = super(Voxels, self).__reduce__()

        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (
            self._number_of_voxels,
            self._voxel_size,
            self._xlim,
            self._ylim,
            self._zlim,
            self._voxel_grids,
        )

        # Return a tuple that replaces the parent's __setstate__ tuple with
        # our own
        return (pickled_state[0], pickled_state[1], new_state)


    def __setstate__(self, state):
        # __reduce__ and __setstate__ ensure correct pickling behaviour
        # https://stackoverflow.com/questions/26598109/preserve-custom-
        # attributes-when-pickling-subclass-of-numpy-array

        # Set the class attributes
        self._voxel_grids = state[-1]
        self._zlim = state[-2]
        self._ylim = state[-3]
        self._xlim = state[-4]
        self._voxel_size = state[-5]
        self._number_of_voxels = state[-6]

        # Call the parent's __setstate__ with the other tuple elements.
        super(Voxels, self).__setstate__(state[0:-6])


    @property
    def voxels(self):
        return self.__array__()


    @property
    def number_of_voxels(self):
        return self._number_of_voxels


    @property
    def xlim(self):
        return self._xlim


    @property
    def ylim(self):
        return self._ylim


    @property
    def zlim(self):
        return self._zlim


    @property
    def voxel_size(self):
        return self._voxel_size


    @property
    def voxel_grids(self):
        return self._voxel_grids


    @staticmethod
    def from_lines(
        lines,
        number_of_voxels,
        xlim = None,
        ylim = None,
        zlim = None,
        verbose = True,
    ):
        '''Create a voxel space and traverse / voxellise a given sample of
        `lines`.

        The `number_of_voxels` in each dimension must be defined. If the
        voxel space boundaries `xlim`, `ylim` or `zlim` are not defined, they
        are inferred as the boundaries of the `lines`.

        Parameters
        ----------
        lines : (M, N>=7) numpy.ndarray or pept.LineData
            The lines that will be voxellised, each defined by a timestamp and
            two 3D points, so that the data columns are [time, x1, y1, z1,
            x2, y2, z2, ...]. Note that extra columns are ignored.

        number_of_voxels : (3,) list[int]
            The number of voxels in the x-, y-, and z-dimensions, respectively.

        xlim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            x-dimension, formatted as [x_min, x_max]. If undefined, it is
            inferred from the boundaries of `lines`.

        ylim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            y-dimension, formatted as [y_min, y_max]. If undefined, it is
            inferred from the boundaries of `lines`.

        zlim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            z-dimension, formatted as [z_min, z_max]. If undefined, it is
            inferred from the boundaries of `lines`.

        Returns
        -------
        pept.Voxels
            A new `Voxels` object with the voxels through which the lines were
            traversed.

        Raises
        ------
        ValueError
            If the input `lines` does not have the shape (M, N>=7). If the
            `number_of_voxels` is not a 1D list with exactly 3 elements, or
            any dimension has fewer than 2 voxels.

        '''
        if verbose:
            start = time.time()

        # Type-checking inputs
        if isinstance(lines, pept.LineData):
            lines = lines.lines

        lines = np.asarray(lines, order = "C", dtype = float)

        if lines.ndim != 2 or lines.shape[1] < 7:
            raise ValueError(textwrap.fill((
                "The input `lines` must be a 2D numpy array containing lines "
                "defined by two points, with every row is formatted as "
                "[time, x1, y1, z1, x2, y2, z2, ...]. The `lines` must then "
                "have shape (M, N >= 7). Received array with shape "
                f"{lines.shape}."
            )))

        number_of_voxels = np.asarray(
            number_of_voxels,
            order = "C",
            dtype = int
        )

        if number_of_voxels.ndim != 1 or len(number_of_voxels) != 3:
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of voxels in the x-, y-, and z-dimension. "
                f"Received parameter with shape {number_of_voxels.shape}."
            )))

        if (number_of_voxels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must set at least two "
                "voxels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_voxels}`."
            )))

        if xlim is None:
            xlim = Voxels.get_cutoff(lines[:, 1], lines[:, 4])
        else:
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the x-dimension. "
                    f"Received parameter with shape {xlim.shape}."
                )))

        if ylim is None:
            ylim = Voxels.get_cutoff(lines[:, 2], lines[:, 5])
        else:
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the y-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        if zlim is None:
            zlim = Voxels.get_cutoff(lines[:, 3], lines[:, 6])
        else:
            zlim = np.asarray(zlim, dtype = float)

            if zlim.ndim != 1 or len(zlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `zlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the z-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        voxels_array = np.zeros(tuple(number_of_voxels))
        voxels = Voxels(
            voxels_array,
            xlim = xlim,
            ylim = ylim,
            zlim = zlim,
        )

        voxels.add_lines(lines, verbose = False)

        if verbose:
            end = time.time()
            print((
                f"Initialised Voxels class in {end - start} s."
            ))

        return voxels


    @staticmethod
    def empty(number_of_voxels, xlim, ylim, zlim):
        '''Create an empty voxel space for the 3D cube bounded by `xlim`,
        `ylim` and `zlim`.

        Parameters
        ----------
        number_of_voxels: (3,) numpy.ndarray
            A list-like containing the number of voxels to be created in the
            x-, y- and z-dimension, respectively.

        xlim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            x-dimension, formatted as [x_min, x_max].

        ylim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            y-dimension, formatted as [y_min, y_max].

        zlim: (2,) numpy.ndarray
            The lower and upper boundaries of the voxellised volume in the
            z-dimension, formatted as [z_min, z_max].

        Raises
        ------
        ValueError
            If `number_of_voxels` does not have exactly 3 values, or it has
            values smaller than 2. If `xlim`, `ylim` or `zlim` do not have
            exactly 2 values each.

        '''

        number_of_voxels = np.asarray(
            number_of_voxels,
            order = "C",
            dtype = int
        )

        if number_of_voxels.ndim != 1 or len(number_of_voxels) != 3:
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of voxels in the x-, y-, and z-dimension. "
                f"Received parameter with shape {number_of_voxels.shape}."
            )))

        if (number_of_voxels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must set at least two "
                "voxels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_voxels}`."
            )))

        number_of_voxels = tuple(number_of_voxels)
        empty_voxels = np.zeros(number_of_voxels)

        return Voxels(
            empty_voxels,
            xlim = xlim,
            ylim = ylim,
            zlim = zlim,
        )


    @staticmethod
    def get_cutoff(p1, p2):
        '''Return a numpy array containing the minimum and maximum value found
        across the two input arrays.

        Parameters
        ----------
        p1 : (N,) numpy.ndarray
            The first 1D numpy array.

        p2 : (N,) numpy.ndarray
            The second 1D numpy array.

        Returns
        -------
        (2,) numpy.ndarray
            The minimum and maximum value found across `p1` and `p2`.

        Notes
        -----
        The input parameters *must* be numpy arrays, otherwise an error will
        be raised.

        '''

        return np.array([
            min(p1.min(), p2.min()),
            max(p1.max(), p2.max()),
        ])


    def save(self, filepath):
        '''Save a `Voxels` instance as a binary `pickle` object.

        Saves the full object state, including the inner `.voxels` NumPy array,
        `xlim`, etc. in a fast, portable binary format. Load back the object
        using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `Voxels` instance, then load it back:

        >>> voxels = pept.Voxels.empty((64, 48, 32), [0, 20], [0, 10], [0, 5])
        >>> voxels.save("voxels.pickle")

        >>> voxels_reloaded = pept.Voxels.load("voxels.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `Voxels` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.Voxels
            The loaded `pept.Voxels` instance.

        Examples
        --------
        Save a `Voxels` instance, then load it back:

        >>> voxels = pept.Voxels.empty((64, 48, 32), [0, 20], [0, 10], [0, 5])
        >>> voxels.save("voxels.pickle")

        >>> voxels_reloaded = pept.Voxels.load("voxels.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def add_lines(self, lines, verbose = False):
        '''Voxellise a sample of lines, adding 1 to each voxel traversed, for
        each line in the sample.

        Parameters
        ----------
        lines : (M, N >= 7) numpy.ndarray
            The sample of 3D lines to voxellise. Each line is defined as a
            timestamp followed by two 3D points, such that the data columns are
            `[time, x1, y1, z1, x2, y2, z2, ...]`. Note that there can be extra
            data columns which will be ignored.

        verbose : bool, default False
            Time the voxel traversal and print it to the terminal.

        Raises
        ------
        ValueError
            If `lines` has fewer than 7 columns.

        '''

        lines = np.asarray(lines, order = "C", dtype = float)
        if lines.ndim != 2 or lines.shape[1] < 7:
            raise ValueError(textwrap.fill((
                "The input `lines` must be a 2D array of lines, where each "
                "line (i.e. row) is defined by a timestamp and two 3D points, "
                "so the data columns are [time, x1, y1, z1, x2, y2, z2]. "
                f"Received array of shape {lines.shape}."
            )))

        if verbose:
            start = time.time()

        traverse3d(
            self.voxels,
            lines,
            self._voxel_grids[0],
            self._voxel_grids[1],
            self._voxel_grids[2]
        )

        if verbose:
            end = time.time()
            print(f"Traversing {len(lines)} lines took {end - start} s.")


    def plot(
        self,
        condition = lambda voxel_data: voxel_data > 0,
        ax = None,
        alt_axes = False,
    ):
        '''Plot the voxels in this class using Matplotlib.

        This plots the centres of all voxels encapsulated in a `pept.Voxels`
        instance, colour-coding the voxel value.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all voxels that have a value larger
        than 0.

        Parameters
        ----------
        condition : function, default `lambda voxel_data: voxel_data > 0`
            The filtering function applied to the voxel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all voxels that
            should be plotted. The default, `lambda x: x > 0` selects all
            voxels which have a value larger than 0.

        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        alt_axes : bool, default False
            If `True`, plot using the alternative PEPT-style axes convention:
            z is horizontal, y points upwards. Because Matplotlib cannot swap
            axes, this is achieved by swapping the parameters in the plotting
            call (i.e. `plt.plot(x, y, z)` -> `plt.plot(z, x, y)`).

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Notes
        -----
        Plotting all points is very computationally-expensive for matplotlib.
        It is recommended to only plot a couple of samples at a time, or use
        the faster `pept.visualisation.PlotlyGrapher`.

        Examples
        --------
        Voxellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> lines = np.array(...)           # shape (N, M >= 7)
        >>> number_of_voxels = [10, 10, 10]
        >>> voxels = pept.Voxels(lines, number_of_voxels)

        >>> fig, ax = voxels.plot()
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection = '3d')
        else:
            fig = plt.gcf()

        filtered_indices = np.argwhere(condition(self))
        positions = self._voxel_size * (0.5 + filtered_indices) + \
            [self._xlim[0], self._ylim[0], self._zlim[0]]

        x = positions[:, 0]
        y = positions[:, 1]
        z = positions[:, 2]

        voxel_vals = np.array([self[tuple(fi)] for fi in filtered_indices])

        cmap = plt.cm.magma
        color_array = cmap(voxel_vals / voxel_vals.max())

        if alt_axes:
            ax.scatter(z, x, y, c = color_array, marker = "s")

            ax.set_xlabel("z (mm)")
            ax.set_ylabel("x (mm)")
            ax.set_zlabel("y (mm)")

        else:
            ax.scatter(x, y, z, c = color_array, marker = "s")

            ax.set_xlabel("x (mm)")
            ax.set_ylabel("y (mm)")
            ax.set_zlabel("z (mm)")

        return fig, ax


    def cube_trace(
        self,
        index,
        color = None,
        opacity = 0.4,
        colorbar = True,
        colorscale = "magma",
    ):
        '''Get the Plotly `Mesh3d` trace for a single voxel at `index`.

        This renders the voxel as a cube. While visually accurate, this method
        is *very* computationally intensive - only use it for fewer than 100
        cubes. For more voxels, use the `voxels_trace` method.

        Parameters
        ----------
        index: (3,) tuple
            The voxel indices, given as a 3-tuple.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.4
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the voxel values. Is overridden if
            `color` is set.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        Raises
        ------
        ValueError
            If `index` does not contain exactly three values.

        Notes
        -----
        If you want to render a small number of voxels as cubes using Plotly,
        use the `cubes_traces` method, which creates a list of individual cubes
        for all voxels, using this function.

        '''

        index = np.asarray(index, dtype = int)

        if index.ndim != 1 or len(index) != 3:
            raise ValueError(textwrap.fill((
                "The input `index` must contain exactly three values, "
                "corresponding to the x, y, z indices of the voxel to plot. "
                f"Received {index}."
            )))

        xyz = self._voxel_size * index + \
            [self._xlim[0], self._ylim[0], self._zlim[0]]

        x = np.array([0, 0, 1, 1, 0, 0, 1, 1]) * self._voxel_size[0]
        y = np.array([0, 1, 1, 0, 0, 1, 1, 0]) * self._voxel_size[1]
        z = np.array([0, 0, 0, 0, 1, 1, 1, 1]) * self._voxel_size[2]
        i = np.array([7, 0, 0, 0, 4, 4, 6, 6, 4, 0, 3, 2])
        j = np.array([3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3])
        k = np.array([0, 7, 2, 3, 6, 7, 1, 1, 5, 5, 7, 6])

        cube = dict(
            x =  x + xyz[0],
            y =  y + xyz[1],
            z =  z + xyz[2],
            i =  i,
            j =  j,
            k =  k,
            opacity = opacity,
            color = color
        )

        if colorbar and color is None:
            cmap = matplotlib.cm.get_cmap(colorscale)
            c = cmap(self[tuple(index)] / (self.max() or 1))
            cube.update(
                color = "rgb({},{},{})".format(c[0], c[1], c[2])
            )

        return go.Mesh3d(cube)


    def cubes_traces(
        self,
        condition = lambda voxels: voxels > 0,
        color = None,
        opacity = 0.4,
        colorbar = True,
        colorscale = "magma",
    ):
        '''Get a list of Plotly `Mesh3d` traces for all voxels selected by the
        `condition` filtering function.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all voxels that have a value larger
        than 0.

        This renders each voxel as individual cubes. While visually accurate,
        this method is *very* computationally intensive - only use it for fewer
        than 100 cubes. For more voxels, use the `voxels_trace` method.

        Parameters
        ----------
        condition : function, default `lambda voxels: voxels > 0`
            The filtering function applied to the voxel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all voxels that
            should be plotted. The default, `lambda x: x > 0` selects all
            voxels which have a value larger than 0.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.4
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the voxel values. Is overridden if
            `color` is set.

        colorscale : str, default "magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        Examples
        --------
        Voxellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines = np.array(...)           # shape (N, M >= 7)

        >>> number_of_voxels = [10, 10, 10]
        >>> voxels = pept.Voxels(lines, number_of_voxels)

        >>> grapher.add_lines(lines)
        >>> grapher.add_traces(voxels.cubes_traces())  # small number of voxels
        >>> grapher.show()

        '''

        indices = np.argwhere(condition(self))
        traces = [
            self.cube_trace(
                i,
                color = color,
                opacity = opacity,
                colorbar = colorbar,
                colorscale = colorscale,
            ) for i in indices
        ]

        return traces


    def voxels_trace(
        self,
        condition = lambda voxel_data: voxel_data > 0,
        size = 4,
        color = None,
        opacity = 0.4,
        colorbar = True,
        colorscale = "Magma",
        colorbar_title = None,
    ):
        '''Create and return a trace for all the voxels in this class, with
        possible filtering.

        Creates a `plotly.graph_objects.Scatter3d` object for the centres of
        all voxels encapsulated in a `pept.Voxels` instance, colour-coding the
        voxel value.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all voxels that have a value larger
        than 0.

        Parameters
        ----------
        condition : function, default `lambda voxel_data: voxel_data > 0`
            The filtering function applied to the voxel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all voxels that
            should be plotted. The default, `lambda x: x > 0` selects all
            voxels which have a value larger than 0.

        size : float, default 4
            The size of the plotted voxel points. Note that due to the large
            number of voxels in typical applications, the *voxel centres* are
            plotted as square points, which provides an easy to understand
            image that is also fast and responsive.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.4
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the voxel values. Is overridden if
            `color` is set.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above it.

        Examples
        --------
        Voxellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines = np.array(...)           # shape (N, M >= 7)
        >>> number_of_voxels = [10, 10, 10]
        >>> voxels = pept.Voxels.from_lines(lines, number_of_voxels)
        >>> grapher.add_lines(lines)
        >>> grapher.add_trace(voxels.voxels_trace())
        >>> grapher.show()

        '''

        filtered_indices = np.argwhere(condition(self))
        positions = self._voxel_size * (0.5 + filtered_indices) + \
            [self._xlim[0], self._ylim[0], self._zlim[0]]

        marker = dict(
            size = size,
            color = color,
            symbol = "square",
        )

        if colorbar and color is None:
            voxel_vals = [self[tuple(fi)] for fi in filtered_indices]
            marker.update(colorscale = "Magma", color = voxel_vals)

            if colorbar_title is not None:
                marker.update(colorbar = dict(title = colorbar_title))

        voxels = dict(
            x = positions[:, 0],
            y = positions[:, 1],
            z = positions[:, 2],
            opacity = opacity,
            mode = "markers",
            marker = marker,
        )

        return go.Scatter3d(voxels)


    def heatmap_trace(
        self,
        ix = None,
        iy = None,
        iz = None,
        width = 0,
        colorscale = "Magma",
        transpose = True
    ):
        '''Create and return a Plotly `Heatmap` trace of a 2D slice through the
        voxels.

        The orientation of the slice is defined by the input `ix` (for the YZ
        plane), `iy` (XZ), `iz` (XY) parameters - which correspond to the
        voxel index in the x-, y-, and z-dimension. Importantly, at least one
        of them must be defined.

        Parameters
        ----------
        ix : int, optional
            The index along the x-axis of the voxels at which a YZ slice is to
            be taken. One of `ix`, `iy` or `iz` must be defined.

        iy: int, optional
            The index along the y-axis of the voxels at which a XZ slice is to
            be taken. One of `ix`, `iy` or `iz` must be defined.

        iz : int, optional
            The index along the z-axis of the voxels at which a XY slice is to
            be taken. One of `ix`, `iy` or `iz` must be defined.

        width : int, default 0
            The number of voxel layers around the given slice index to collapse
            (i.e. accumulate) onto the heatmap.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        transpose : bool, default True
            Transpose the heatmap (i.e. flip it across its diagonal).

        Raises
        ------
        ValueError
            If neither of `ix`, `iy` or `iz` was defined.

        Examples
        --------
        Voxellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> lines = np.array(...)           # shape (N, M >= 7)
        >>> number_of_voxels = [10, 10, 10]
        >>> voxels = pept.Voxels(lines, number_of_voxels)

        >>> import plotly.graph_objs as go
        >>> fig = go.Figure()
        >>> fig.add_trace(voxels.heatmap_trace())
        >>> fig.show()

        '''

        if ix is not None:
            x = self._voxel_grids[1]
            y = self._voxel_grids[2]
            z = self[ix, :, :]

            for i in range(1, width + 1):
                z = z + self[ix + i, :, :]
                z = z + self[ix - i, :, :]

        elif iy is not None:
            x = self._voxel_grids[0]
            y = self._voxel_grids[2]
            z = self[:, iy, :]

            for i in range(1, width + 1):
                z = z + self[:, iy + i, :]
                z = z + self[:, iy - i, :]

        elif iz is not None:
            x = self._voxel_grids[0]
            y = self._voxel_grids[1]
            z = self[:, :, iz]

            for i in range(1, width + 1):
                z = z + self[:, :, iz + i]
                z = z + self[:, :, iz - i]

        else:
            raise ValueError(textwrap.fill((
                "[ERROR]: One of the `ix`, `iy`, `iz` slice indices must be "
                "provided."
            )))

        heatmap = dict(
            x = x,
            y = y,
            z = z,
            colorscale = colorscale,
            transpose = transpose,
        )

        return go.Heatmap(heatmap)


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"{self.__array__()}\n\n"
            f"number_of_voxels =    {self._number_of_voxels}\n"
            f"voxel_size =          {self._voxel_size}\n\n"
            f"xlim =                {self._xlim}\n"
            f"ylim =                {self._ylim}\n"
            f"zlim =                {self._zlim}\n\n"
            f"voxel_grids:\n"
            f"([{self._voxel_grids[0][0]} ... {self._voxel_grids[0][-1]}],\n"
            f" [{self._voxel_grids[1][0]} ... {self._voxel_grids[1][-1]}],\n"
            f" [{self._voxel_grids[2][0]} ... {self._voxel_grids[2][-1]}])"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPR
        docstr = (
            "Class instance that inherits from `pept.Voxels`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}"
        )

        return docstr




class VoxelData:
    '''A class that can voxellise multiple samples of lines (from a `LineData`)
    lazily / on demand.

    Voxellisation is a computationally-intensive step - in terms of both time
    and memory - especially for many thousands of lines or samples; although
    optimised, the `Voxels` class only manages a single voxel space. The simple
    solution for multiple samples:

    .. code-block:: python

        lines = pept.LineData(...)
        voxels = []

        for sample in lines:
            voxels.append( pept.Voxels(sample, (100, 100, 100)) )


    Is very inefficient, as it uses a single thread and stores all voxel
    spaces in memory (voxels use up a lot of memory!). This class solves these
    problems by:

    1. Voxellising the samples of lines in parallel, on any number of threads.
    2. Creating the voxel spaces on demand. Optionally, it can save / cache
       the expensive voxellisation steps (`save_cache = True`).

    The individual voxellisation steps are still done using `Voxels`, which are
    then accessible in a list (`voxels.voxels` or `voxels[0]`).

    Attributes
    ----------
    line_data : pept.LineData
        The samples of lines encapsulated in a `pept.LineData` that will be
        used to create the corresponding voxellised spaces.

    voxels : list[pept.Voxels | None]
        The list of cached voxellised spaces. Initially, it is a list of
        `None`s of the same length as the number of samples in `line_data`. If
        `save_cache` is True, the voxellised samples of lines will be cached
        here.

    number_of_voxels : 3-tuple
        A 3-tuple corresponding to the shape of `voxels`.

    xlim : (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        x-dimension, formatted as [x_min, x_max].

    ylim : (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        y-dimension, formatted as [y_min, y_max].

    zlim : (2,) numpy.ndarray
        The lower and upper boundaries of the voxellised volume in the
        z-dimension, formatted as [z_min, z_max].

    max_workers : int
        The number of threads that will be used to voxellise the lines in
        parallel.

    save_cache : bool
        Whether to cache the voxellised spaces *when their computation is
        requested* (e.g. when calling `traverse()`). The `voxels` attribute is
        only populated if this is `True`.

    Methods
    -------
    save(filepath)
        Save a `VoxelData` instance as a binary `pickle` object.

    load(filepath)
        Load a saved / pickled `VoxelData` object from `filepath`.

    traverse(sample_indices = ..., verbose = True)
        Voxellise the samples in `line_data` at indices `samples_indices`.

    accumulate(sample_indices = ..., verbose = True):
        Superimpose the voxellised samples in `line_data` at indices
        `samples_indices` into the same voxel space.

    Notes
    -----
    Upon instantiating the class with an instance of `LineData`, a copy is
    made. It is a logic error to change the `sample_size` and `overlap` after
    instantiation; these should remain constant for the `voxels` list to
    remain correct.

    Examples
    --------
    Create a short list of 3D lines and encapsulate them in a `LineData` class
    to simulate multiple samples of lines.

    >>> import pept
    >>> import numpy as np

    >>> lines_raw = np.arange(70).reshape(10, 7)
    >>> lines = pept.LineData(lines_raw, sample_size = 2)

    The `VoxelData` does not voxellise the samples of lines immediately upon
    instantiation by default:

    >>> number_of_voxels = [3, 4, 5]
    >>> voxel_data = pept.VoxelData(lines, number_of_voxels)
    >>> voxel_data.voxels
    >>> [None, None, None, None, None]

    You can iterate through the `VoxelData` and it will voxellise the samples
    on demand:

    >>> for vox in voxel_data:
    >>>     print(vox)          # not cached by default

    This is the most efficient way to use `VoxelData`, as the voxellised
    samples are NOT cached by default; they are deleted after each loop above
    (voxels consume a lot of memory!):

    >>> voxel_data.voxels
    >>> [None, None, None, None, None]

    If you have enough memory to store all the voxel spaces at once, set
    `save_cache` to True when instantiating the class, or afterwards:

    >>> number_of_voxels = [3, 4, 5]
    >>> voxel_data = pept.VoxelData(lines, number_of_voxels, save_cache = True)
    >>> voxel_data.voxels
    >>> [None, None, None, None, None]

    >>> voxel_data.traverse()               # Actually voxellises each sample
    >>> voxel_data.voxels
    >>> [..Voxels.. , ..Voxels.., ...]

    If you voxellised the samples once and `save_cache = True`, the results are
    cached, so new iterations use the pre-computed voxel spaces:

    >>> voxel_data.accumulate()             # Almost instantaneous, uses cache

    See Also
    --------
    pept.Voxels : Manage a single voxellised sample of lines.
    pept.LineData : Encapsulate lines for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.

    '''

    def __init__(
        self,
        line_data,
        number_of_voxels,
        xlim = None,
        ylim = None,
        zlim = None,
        max_workers = None,
        save_cache = False,
        traverse = False,
        verbose = True,
    ):
        '''`VoxelData` class constructor.

        Parameters
        ----------
        line_data : pept.LineData
            The samples of lines encapsulated in a `pept.LineData` that will be
            used to create the corresponding voxellised spaces.

        number_of_voxels : (3,) list[int]
            The number of voxels in the x-, y-, and z-dimensions, respectively.

        xlim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            x-dimension, formatted as [x_min, x_max]. If undefined, it is
            inferred from the boundaries of the lines in `line_data`.

        ylim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            y-dimension, formatted as [y_min, y_max]. If undefined, it is
            inferred from the boundaries of the lines in `line_data`.

        zlim : (2,) list[float], optional
            The lower and upper boundaries of the voxellised volume in the
            z-dimension, formatted as [z_min, z_max]. If undefined, it is
            inferred from the boundaries of the lines in `line_data`.

        max_workers : int, optional
            The number of threads that will be used to voxellise the lines in
            parallel. If undefined, the value returned by `os.cpu_count()` is
            used.

        save_cache : bool, default False
            Whether to cache the voxellised spaces *when their computation is
            requested* (e.g. when calling `traverse()`). The `voxels` attribute
            is only populated if this is `True`.

        traverse : bool, default False
            Whether to immediately voxellise all the samples of lines in
            `line_data`. If `True`, `save_cache` is also set to `True` and the
            `traverse()` method is called.

        verbose : bool, default True
            Show extra information as the voxellisation runs.

        Raises
        ------
        TypeError
            If `line_data` is not an instance (or subclass) of `pept.LineData`.

        ValueError
            If `number_of_voxels` does not have exactly 3 values, or it has
            values smaller than 2. If `xlim`, `ylim` or `zlim` do not have
            exactly 2 values each.

        '''

        # Type-checking inputs
        if not isinstance(line_data, pept.LineData):
            raise TypeError(textwrap.fill((
                "The input `line_data` must be an instance (or subclass "
                f"thereof!) of `pept.LineData`. Received {type(line_data)}."
            )))

        number_of_voxels = np.asarray(
            number_of_voxels,
            order = "C",
            dtype = int
        )

        if number_of_voxels.ndim != 1 or len(number_of_voxels) != 3:
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of voxels in the x-, y-, and z-dimension. "
                f"Received parameter with shape {number_of_voxels.shape}."
            )))

        if (number_of_voxels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must set at least two "
                "voxels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_voxels}`."
            )))

        # Alias for the whole array of lines in `line_data`
        lines = line_data.lines

        if xlim is None:
            xlim = Voxels.get_cutoff(lines[:, 1], lines[:, 4])
        else:
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the x-dimension. "
                    f"Received parameter with shape {xlim.shape}."
                )))

        if ylim is None:
            ylim = Voxels.get_cutoff(lines[:, 2], lines[:, 5])
        else:
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the y-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        if zlim is None:
            zlim = Voxels.get_cutoff(lines[:, 3], lines[:, 6])
        else:
            zlim = np.asarray(zlim, dtype = float)

            if zlim.ndim != 1 or len(zlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `zlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the z-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        save_cache = bool(save_cache)
        traverse = bool(traverse)

        # Setting class attributes
        self._line_data = line_data.copy()
        self._number_of_voxels = tuple(number_of_voxels)
        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim

        self._max_workers = max_workers
        self._save_cache = save_cache

        self._voxels_index = 0
        self._voxels = [None for _ in range(len(self._line_data))]

        if traverse:
            self._save_cache = True
            self.traverse(verbose = verbose)


    @property
    def line_data(self):
        return self._line_data


    @property
    def voxels(self):
        return self._voxels


    @property
    def number_of_voxels(self):
        return self._number_of_voxels


    @property
    def xlim(self):
        return self._xlim


    @property
    def ylim(self):
        return self._ylim


    @property
    def zlim(self):
        return self._zlim


    @property
    def max_workers(self):
        return self._max_workers


    @property
    def save_cache(self):
        return self._save_cache


    @save_cache.setter
    def save_cache(self, new_save_cache):
        self._save_cache = bool(new_save_cache)


    def save(self, filepath):
        '''Save a `VoxelData` instance as a binary `pickle` object.

        Saves the full object state, including the inner `.voxels` NumPy array,
        `xlim`, etc. in a fast, portable binary format. Load back the object
        using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `VoxelData` instance (created from lines), then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> voxel_data = pept.VoxelData(lines, (64, 48, 32))
        >>> voxel_data.save("voxel_data.pickle")

        >>> voxel_data_reloaded = pept.VoxelData.load("voxel_data.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `VoxelData` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.VoxelData
            The loaded `pept.VoxelData` instance.

        Examples
        --------
        Save a `VoxelData` instance (created from lines), then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> voxel_data = pept.VoxelData(lines, (64, 48, 32))
        >>> voxel_data.save("voxel_data.pickle")

        >>> voxel_data_reloaded = pept.VoxelData.load("voxel_data.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def traverse(
        self,
        sample_indices = ...,
        verbose = True,
    ):
        '''Voxellise the samples in `line_data` at indices `samples_indices`.

        If `save_cache` is `True`, the voxellised spaces are also cached in the
        `voxels` attribute. Otherwise, they are only returned as a list.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of lines. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all samples).

        verbose : bool, default True
            Show extra information as the voxellisation runs.

        Returns
        -------
        list[Voxels]
            A list of the voxellised samples of lines (each as a `Voxels`
            class) selected by `sample_indices`.

        Examples
        --------
        This method is automatically called if the instantiation of the class
        sets `traverse = True`:

        >>> import pept
        >>> import numpy as np

        >>> lines_raw = np.arange(70).reshape(10, 7)
        >>> lines = pept.LineData(lines_raw, sample_size = 2)

        >>> number_of_voxels = [3, 4, 5]
        >>> voxel_data = pept.VoxelData(
        >>>     lines, number_of_voxels, traverse = True
        >>> )
        >>> voxel_data.voxels
        >>> [..Voxels.. , ..Voxels.., ...]

        You can also call the method after the instantiation. If the results
        were cached (i.e. `save_cache = True`), they are returned without
        re-computing them:

        >>> voxel_list = voxel_data.traverse()

        Alternatively, you can only voxellise a few samples at once (still in
        parallel):

        >>> voxel_list = voxel_data.traverse([1, 5, 2])

        Of course, the list of `Voxels` will have the same ordering as the
        list of indices given to the function.

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis.
        if sample_indices is Ellipsis:
            sample_indices = np.arange(len(self._line_data))
        elif not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        with ThreadPoolExecutor(max_workers = self._max_workers) as executor:
            # Use pre-computed voxels and voxellise the other samples
            selected_voxels = [None for _ in range(len(sample_indices))]
            selected_futures = [None for _ in range(len(sample_indices))]

            # Voxellise each selected sample
            for i, n in enumerate(sample_indices):
                # Optimisation: if this sample was already voxellised, reuse it
                if self._voxels[n] is not None:
                    selected_voxels[i] = self._voxels[n].copy()
                    continue

                # Otherwise, voxellise the sample asynchronously
                selected_futures[i] = executor.submit(
                    pept.Voxels.from_lines,
                    self._line_data[n],
                    self._number_of_voxels,
                    xlim = self._xlim,
                    ylim = self._ylim,
                    zlim = self._zlim,
                    verbose = False,
                )

            if verbose:
                sample_indices = tqdm(sample_indices)

            # Iterate through all the futures; if not None (i.e. we voxellised
            # the sample now), get the result. Otherwise it was pre-computed
            for i, n in enumerate(sample_indices):
                if selected_futures[i] is not None:
                    selected_voxels[i] = selected_futures[i].result()

                    # If we voxellised this sample and we use "save_cache",
                    # cache the result in self.voxels
                    if self._save_cache:
                        self._voxels[n] = selected_voxels[i]

        return selected_voxels


    def accumulate(self, sample_indices = ..., verbose = True):
        '''Superimpose the voxellised samples in `line_data` at indices
        `samples_indices` into the same voxel space.

        This method can be used to voxellise multiple samples of lines into the
        same `Voxels` class. The computation is done in parallel and uses the
        least amount of memory possible.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of lines. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all samples).

        verbose : bool, default True
            Show extra information as the voxellisation runs.

        Returns
        -------
        Voxels
            The single voxellised space that contains the superimposed
            voxellised representations of the samples of lines selected by
            `sample_indices`.

        Examples
        --------
        This method is helpful for splitting the voxellisation that would be
        done by a single `Voxels` class across multiple threads:

        >>> import pept
        >>> import numpy as np

        >>> lines_raw = np.arange(70).reshape(10, 7)
        >>> lines = pept.LineData(lines_raw, sample_size = 2)

        >>> number_of_voxels = [3, 4, 5]
        >>> voxel_data = pept.VoxelData(lines, number_of_voxels)
        >>> vox = voxel_data.accumulate()

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis.
        if sample_indices is Ellipsis:
            sample_indices = np.arange(len(self._line_data))
        elif not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        vox = Voxels.empty(
            self._number_of_voxels,
            self._xlim,
            self._ylim,
            self._zlim,
        )

        # Voxellise each selected sample
        with ThreadPoolExecutor(max_workers = self._max_workers) as executor:
            # Use pre-computed voxels and voxellise the other samples
            selected_futures = [None for _ in range(len(sample_indices))]

            for i, n in enumerate(sample_indices):
                # Optimisation: if this sample was already voxellised, reuse it
                if self._voxels[n] is not None:
                    continue

                # Otherwise, voxellise the sample asynchronously
                selected_futures[i] = executor.submit(
                    pept.Voxels.from_lines,
                    self._line_data[n],
                    self._number_of_voxels,
                    xlim = self._xlim,
                    ylim = self._ylim,
                    zlim = self._zlim,
                    verbose = False,
                )

            if verbose:
                sample_indices = tqdm(sample_indices)

            # Iterate through all the futures; if not None (i.e. we voxellised
            # the sample now), get the result. Otherwise it was pre-computed
            for i, n in enumerate(sample_indices):
                if selected_futures[i] is not None:
                    vox += selected_futures[i].result()
                else:
                    vox += self._voxels[n]

        return vox


    def __getitem__(self, key):
        # For accessing voxels using subscript notation
        key = int(key)

        # Allow negative indices
        while key < 0:
            key += len(self._voxels)

        if key >= len(self._voxels):
            raise IndexError(textwrap.fill((
                f"The index `{key}` was out of range. There are "
                f"{len(self._voxels)} samples of lines to be voxellised, "
                "indexed from 0."
            )))

        if self._voxels[key] is not None:
            return self._voxels[key]

        vox = pept.Voxels.from_lines(
            self._line_data[key],
            self._number_of_voxels,
            xlim = self._xlim,
            ylim = self._ylim,
            zlim = self._zlim,
            verbose = False,
        )

        if self._save_cache:
            self._voxels[key] = vox

        return vox


    def __iter__(self):
        return self


    def __next__(self):
        # Finish iteration
        if self._voxels_index >= len(self._line_data):
            self._voxels_index = 0

            raise StopIteration

        # If we already traversed and saved this sample, return it directly
        if self._voxels[self._voxels_index] is not None:
            self._voxels_index += 1
            return self._voxels[self._voxels_index - 1]

        vox = pept.Voxels.from_lines(
            self._line_data[self._voxels_index],
            self._number_of_voxels,
            xlim = self._xlim,
            ylim = self._ylim,
            zlim = self._zlim,
            verbose = False,
        )

        if self._save_cache:
            self._voxels[self._voxels_index] = vox

        self._voxels_index += 1

        return vox


    def __len__(self):
        return len(self._voxels)


    def __str__(self):
        # Shown when calling print(class)
        traversed = sum((v is not None for v in self._voxels))
        total = len(self._voxels)

        # Indent LineData docstring with an extra tab
        line_data_str = "    ".join(self._line_data.__str__().splitlines(True))

        docstr = (
            f"voxels:               {traversed} / {total} traversed & cached\n"
            f"number_of_voxels =    {self._number_of_voxels}\n"
            f"xlim =                {self._xlim}\n"
            f"ylim =                {self._ylim}\n"
            f"zlim =                {self._zlim}\n\n"
            f"line_data:\n    {line_data_str}\n\n"
            f"save_cache =          {self._save_cache}\n"
            f"max_workers =         {self._max_workers}\n"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPR
        docstr = (
            "Class instance that inherits from `pept.VoxelData`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}"
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : peptml.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 28.08.2019


import copy
import os
import sys
import textwrap
import time
import warnings

import  numpy               as      np

from    joblib              import  Parallel, delayed
from    tqdm                import  tqdm

import  cma

# Fix a deprecation warning inside the sklearn library
try:
    sys.modules['sklearn.externals.six'] = __import__('six')
    sys.modules['sklearn.externals.joblib'] = __import__('joblib')
    import hdbscan
except ImportError:
    import hdbscan

import  pept




class HDBSCANClusterer:
    '''Efficient, optionally-parallel HDBSCAN-based clustering for cutpoints
    computed from LoRs (or generic 3D points).

    Given a sample of (cut)points, they are clustered using the HDBSCAN [1]_
    algorithm; the tracers' locations are computed as the "centre" of each
    cluster, corresponding to the average of all the points in the cluster
    (this is the original implementation [2]_). Optionally, the cluster
    exemplars - that is, the most stable points in the cluster - can be used
    for the average (if `select_exemplars` == True).

    This algorithm requires no prior knowledge of the number of tracers within
    the system; it successfully distinguished multiple particles separated by
    distances as small as 2 mm. The technique’s spatial resolution was observed
    to be invariant with the number of tracers used, allowing large numbers of
    particles to be tracked simultaneously, with no loss of data quality [2]_.

    This class is a wrapper around the `hdbscan` package, providing tools for
    parallel clustering of samples of cutpoints. It can return `PointData`
    classes for ease of manipulation and visualisation.

    Two main methods are provided: `fit_sample` for clustering a single numpy
    array of cutpoints (i.e. a single sample) and `fit` which clusters all the
    samples encapsulated in a `pept.PointData` class (such as the one returned
    by the `Cutpoints` class) *in parallel*.

    Optionally, this class can find the optimum clustering parameters for a
    given dataset using the `optimise` method.

    Attributes
    ----------
    clusterer : hdbscan.HDBSCAN object
        The HDBSCAN object that will be used to cluster points. The settings
        below apply to it. If `allow_single_cluster` is set to the default
        "auto", a second clusterer object (`clusterer_single`) will be used for
        samples in which no cluster was found. This is useful for experiments
        in which tracers leave and enter the field of view, or coalesce. The
        `optimise` method can find the optimal settings for a given dataset.

    clusterer_single : hdbscan.HDBSCAN object or None
        The second HDBSCAN object, defined only if `allow_single_cluster` is
        "auto". It can have different settings than `clusterer`, which is
        useful for long experiments in which the data looks very different
        when a single tracer remains in the field of view (e.g. if two tracers
        coalesce, the resulting "single tracer" will be much more active). The
        `optimise` method can find the optimal settings for a given dataset,
        independently from `clusterer`.

    min_cluster_size : int
        (Taken from hdbscan's documentation): The minimum size of clusters;
        single linkage splits that contain fewer points than this will be
        considered points “falling out” of a cluster rather than a cluster
        splitting into two new clusters.

    min_samples : int
        (Taken from hdbscan's documentation): The number of samples in a
        neighbourhood for a point to be considered a core point. The default is
        `None`, being set automatically to the `min_cluster_size`.

    min_cluster_size_single : int or None
        The `min_cluster_size` value for `clusterer_single`, defined only if
        `allow_single_cluster` is "auto".

    min_samples_single : int
        The `min_samples` value for `clusterer_single`, defined only if
        `allow_single_cluster` is "auto".

    allow_single_cluster : bool or "auto"
        By default HDBSCAN will not produce a single cluster - this creates
        "tighter" clusters (i.e. more accurate positions). Setting this to
        `False` will discard datasets with a single tracer, but will produce
        more accurate positions. Setting this to `True` will also work for
        single tracers, at the expense of lower accuracy for cases with more
        tracers. This class provides a third option, "auto", in which case two
        clusterers will be used: one with `allow_single_cluster` set to `False`
        and another with it set to `True`; the latter will only be used if the
        first did not find any clusters.

    select_exemplars : bool, default False
        If `False`, the tracer position is computed as the average of *all* the
        points in a cluster. If `True`, the tracer position is computed as the
        average of the *cluster exemplars*. This can yield tighter clusters in
        the majority of cases, but miss some tracers if the points are very
        sparse.

    labels : (N,) numpy.ndarray, dtype = int
        A 1D array of the cluster labels for cutpoints fitted using
        `fit_sample` or `fit`. If `fit_sample` is used, `labels` correspond to
        each row in the sample array fitted. If `fit` is used with the setting
        `get_labels = True`, `labels` correpond to the stacked labels for every
        sample in the given `pept.PointData` class.

    Methods
    -------
    fit_sample(sample, get_labels = False, as_array = True, verbose = False,\
               _set_labels = True)
        Fit one sample of cutpoints and return the cluster centres and
        (optionally) the labelled cutpoints.

    fit(cutpoints, get_labels = False, max_workers = None, verbose = True)
        Fit cutpoints (an instance of `PointData`) and return the cluster
        centres and (optionally) the labelled cutpoints.

    optimise(points, nsamples = 16, max_workers = None, verbose = True,\
             _stability_params = None)
        Optimise HDBSCANClusterer settings against a given dataset of `points`.

    Examples
    --------
    A typical workflow would involve reading LoRs from a file, computing their
    cutpoints, clustering them and plotting them.

    >>> import pept
    >>> from pept.tracking import peptml
    >>>
    >>> lors = pept.LineData(...)
    >>> cutpoints = peptml.Cutpoints(lors, 0.15)
    >>>
    >>> clusterer = peptml.HDBSCANClusterer()
    >>> clusterer.optimise(cutpoints)           # optional
    >>> centres = clusterer.fit(cutpoints)
    >>>
    >>> grapher = PlotlyGrapher()
    >>> grapher.add_points(centres)
    >>> grapher.show()

    For more advanced uses of HDBSCANClusterer such as 2-pass clustering, do
    check out the tutorials available on the Birmingham's Positron Imaging
    Centre's GitHub repository at github.com/uob-positron-imaging-centre.

    References
    ----------
    .. [1] McInnes L, Healy J. Accelerated Hierarchical Density Based
       Clustering In: 2017 IEEE International Conference on Data Mining
       Workshops (ICDMW), IEEE, pp 33-42. 2017
    .. [2] Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
       using machine learning. Review of Scientific Instruments. 2020 Jan 1;
       91(1):013329.

    See Also
    --------
    pept.tracking.peptml.Cutpoints : Compute cutpoints from `pept.LineData`.
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    def __init__(
        self,
        min_cluster_size = 20,
        min_samples = None,
        min_cluster_size_single = None,
        min_samples_single = None,
        allow_single_cluster = "auto",
        select_exemplars = False,
        core_dist_n_jobs = 1,
        **kwargs
    ):
        '''`HDBSCANClusterer` class constructor.

        Parameters
        ----------
        min_cluster_size : int, default 20
            (Taken from hdbscan's documentation): The minimum size of clusters;
            single linkage splits that contain fewer points than this will be
            considered points “falling out” of a cluster rather than a cluster
            splitting into two new clusters.

        min_samples : int, optional
            (Taken from hdbscan's documentation): The number of samples in a
            neighbourhood for a point to be considered a core point. The
            default is `None`, being set automatically to the
            `min_cluster_size`.

        min_cluster_size_single : int, optional
            The `min_cluster_size` value for `clusterer_single`, defined only
            if `allow_single_cluster` is "auto". The default is `None`, being
            set automatically to the `min_cluster_size`.

        min_samples_single : int, optional
            The `min_samples` value for `clusterer_single`, defined only if
            `allow_single_cluster` is "auto". The default is `None`, being set
            automatically to the `min_cluster_size`.

        allow_single_cluster : bool or str, default "auto"
            By default HDBSCAN will not produce a single cluster - this creates
            "tighter" clusters (i.e. more accurate positions). Setting this to
            `False` will discard datasets with a single tracer, but will
            produce more accurate positions. Setting this to `True` will also
            work for single tracers, at the expense of lower accuracy for cases
            with more tracers. This class provides a third option, "auto", in
            which case two clusterers will be used: one with
            `allow_single_cluster` set to `False` and another with it set to
            `True`; the latter will only be used if the first did not find any
            clusters.

        select_exemplars : bool, default False
            If `False`, the tracer position is computed as the average of *all*
            the points in a cluster. If `True`, the tracer position is computed
            as the average of the *cluster exemplars*. This can yield tighter
            clusters in the majority of cases, but miss some tracers if the
            points are very sparse.

        core_dist_n_jobs : int or None, default 1
            (Taken from hdbscan's documentation): Number of parallel jobs to
            run in core distance computations (if supported by the specific
            algorithm). If unset (`None`), the value returned by
            `os.cpu_count()` is used. This should be left to the default 1, as
            parallelism is more advantageous between samples (i.e. when calling
            `fit`) rather than for each clustering algorithm run.

        kwargs : keyword arguments
            Other keyword arguments that will be passed to the HDBSCAN
            instantiation.

        Raises
        ------
        ValueError
            If `allow_single_cluster` is not `True`, `False` or "auto".
        '''

        # Type-check min_cluster_size and min_samples
        min_cluster_size = int(round(min_cluster_size))

        if min_samples is None:
            min_samples = min_cluster_size
        else:
            min_samples = int(round(min_samples))

        if min_cluster_size_single is None:
            min_cluster_size_single = min_cluster_size
        else:
            min_cluster_size_single = int(round(min_cluster_size_single))

        if min_samples_single is None:
            min_samples_single = min_samples
        else:
            min_samples_single = int(round(min_samples_single))

        if core_dist_n_jobs is None:
            core_dist_n_jobs = os.cpu_count()

        self._select_exemplars = bool(select_exemplars)
        self._labels = None

        # Will manually supply allow_single_cluster. Only need prediction_data
        # if select_exemplars is `True`.
        clusterer_options = {
            "min_cluster_size": min_cluster_size,
            "min_samples": min_samples,
            "core_dist_n_jobs": core_dist_n_jobs,
            "gen_min_span_tree": True,
            "prediction_data": self._select_exemplars,
            "cluster_selection_method": "leaf",         # Quite harsh
            **kwargs
        }

        clusterer_single_options = {
            "min_cluster_size": min_cluster_size_single,
            "min_samples": min_samples_single,
            "core_dist_n_jobs": core_dist_n_jobs,
            "gen_min_span_tree": True,
            "prediction_data": self._select_exemplars,
            "cluster_selection_method": "eom",          # More lenient
            **kwargs
        }

        # Create inner clusterers (including clusterer_single if
        # allow_single_cluster == "auto")
        if allow_single_cluster is True:
            self._allow_single_cluster = allow_single_cluster
            self.clusterer = hdbscan.HDBSCAN(
                allow_single_cluster = True,
                **clusterer_options
            )

            self.clusterer_single = None

        elif allow_single_cluster is False:
            self._allow_single_cluster = allow_single_cluster
            self.clusterer = hdbscan.HDBSCAN(
                allow_single_cluster = False,
                **clusterer_options
            )

            self.clusterer_single = None

        elif str(allow_single_cluster).lower() == "auto":
            self._allow_single_cluster = "auto"
            self.clusterer = hdbscan.HDBSCAN(
                allow_single_cluster = False,
                **clusterer_options
            )

            self.clusterer_single = hdbscan.HDBSCAN(
                allow_single_cluster = True,
                **clusterer_single_options
            )

        else:
            raise ValueError((
                "\n[ERROR]: `allow_single_cluster` should be either `True`, "
                f"`False` or 'auto'. Received {allow_single_cluster}.\n"
            ))


    @property
    def min_cluster_size(self):
        return self.clusterer.min_cluster_size


    @min_cluster_size.setter
    def min_cluster_size(self, new_min_cluster_size):
        if new_min_cluster_size < 2:
            warnings.warn((
                "\n[WARNING]: new_min_cluster_size was set to 2, as it was "
                f"{new_min_cluster_size} < 2.\n"
            ), RuntimeWarning)
            new_min_cluster_size = 2
        else:
            new_min_cluster_size = int(round(new_min_cluster_size))

        self.clusterer.min_cluster_size = new_min_cluster_size


    @property
    def min_samples(self):
        return self.clusterer.min_samples


    @min_samples.setter
    def min_samples(self, new_min_samples):
        if new_min_samples < 2:
            warnings.warn((
                "\n[WARNING]: new_min_samples was set to 2, as it was "
                f"{new_min_samples} < 2.\n"
            ), RuntimeWarning)
            new_min_samples = 2
        else:
            new_min_samples = int(round(new_min_samples))

        self.clusterer.min_samples = new_min_samples


    @property
    def min_cluster_size_single(self):
        if self.clusterer_single is None:
            return None

        return self.clusterer_single.min_cluster_size


    @min_cluster_size_single.setter
    def min_cluster_size_single(self, new_min_cluster_size):
        if new_min_cluster_size < 2:
            warnings.warn((
                "\n[WARNING]: new_min_cluster_size was set to 2, as it was "
                f"{new_min_cluster_size} < 2.\n"
            ), RuntimeWarning)
            new_min_cluster_size = 2
        else:
            new_min_cluster_size = int(round(new_min_cluster_size))

        self.clusterer_single.min_cluster_size = new_min_cluster_size


    @property
    def min_samples_single(self):
        if self.clusterer_single is None:
            return None

        return self.clusterer_single.min_samples


    @min_samples_single.setter
    def min_samples_single(self, new_min_samples):
        if new_min_samples < 2:
            warnings.warn((
                "\n[WARNING]: new_min_samples was set to 2, as it was "
                f"{new_min_samples} < 2.\n"
            ), RuntimeWarning)
            new_min_samples = 2
        else:
            new_min_samples = int(round(new_min_samples))

        self.clusterer_single.min_samples = new_min_samples


    @property
    def allow_single_cluster(self):
        return self._allow_single_cluster


    @allow_single_cluster.setter
    def allow_single_cluster(self, allow_single_cluster):
        if allow_single_cluster is True:
            self._allow_single_cluster = allow_single_cluster
            self.clusterer.allow_single_cluster = True
            self.clusterer_single = None

        elif allow_single_cluster is False:
            self._allow_single_cluster = allow_single_cluster
            self.clusterer.allow_single_cluster = False
            self.clusterer_single = None

        elif str(allow_single_cluster).lower() == "auto":
            self._allow_single_cluster = "auto"
            self.clusterer.allow_single_cluster = False

            # Make a deep copy of the "normal" clusterer, and only change the
            # `allow_single_cluster` option to `True`.
            self.clusterer_single = copy.deepcopy(self.clusterer)
            self.clusterer_single.allow_single_cluster = True

        else:
            raise ValueError((
                "\n[ERROR]: `allow_single_cluster` should be either `True`, "
                f"`False` or 'auto'. Received {allow_single_cluster}.\n"
            ))


    @property
    def select_exemplars(self):
        return self._select_exemplars


    @select_exemplars.setter
    def select_exemplars(self, select_exemplars):
        self._select_exemplars = bool(select_exemplars)

        self.clusterer.prediction_data = self._select_exemplars

        if self.clusterer_single is not None:
            self.clusterer_single.prediction_data = self._select_exemplars


    @property
    def labels(self):
        return self._labels


    def fit_sample(
        self,
        sample,
        get_labels = False,
        as_array = True,
        verbose = False,
        _set_labels = True
    ):
        '''Fit one sample of cutpoints and return the cluster centres and
        (optionally) the labelled cutpoints.

        Parameters
        ----------
        sample : (N, M >= 4) numpy.ndarray
            The sample of points that will be clustered. The expected columns
            are `[time, x, y, z, etc]`. Only columns `[1, 2, 3]` are used for
            clustering.

        get_labels : bool, default False
            If set to True, the input `sample` is also returned with an extra
            column representing the label of the cluster that each point is
            associated with. This label is an `int`, numbering clusters
            starting from 0; noise is represented with the value -1.

        as_array : bool, default True
            If set to True, the centres of the clusters and the labelled
            cutpoints are returned as numpy arrays. If set to False, they are
            returned inside instances of `pept.PointData`.

        verbose : bool, default False
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        _set_labels : bool, default True
            This is an internal setting that an end-user should not normally
            care about. If `True`, the class property `labels` will be set
            after fitting. Setting this to `False` is helpful for multithreaded
            contexts - when calling `fit_sample` in parallel, it makes sure
            no internal attributes are mutated at the same time.

        Returns
        -------
        centres : numpy.ndarray or pept.PointData
            The centroids of every cluster found with columns
            `[time, x, y, z, ..., cluster_size]`. They are computed as the
            column-wise average of the points included in each cluster (i.e.
            for each label) or the cluster exemplars (if the `select_exemplars`
            attribute is `True`). Another column is added to the initial data
            in `sample`, signifying the cluster size - that is, the number of
            points included in the cluster. If `as_array` is set to True, it is
            a numpy array, otherwise the centres are stored in a
            `pept.PointData` instance.

        sample_labelled : optional, numpy.ndarray or pept.PointData
            Returned if `get_labels` is `True`. It is the input `sample` with
            an appended column representing the label of the cluster that the
            point was associated with. The labels are integers starting from 0.
            The points classified as noise have the number -1 associated. If
            `as_array` is set to True, it is a numpy array, otherwise the
            labelled points are stored in a `pept.PointData` instance.

        Raises
        ------
        ValueError
            If `sample` is not a numpy array of shape (N, M), where M >= 4.

        Notes
        -----
        If no clusters were found (i.e. all labels are -1), the returned values
        are empty numpy arrays.
        '''

        if verbose:
            start = time.time()

        # sample columns: [time, x, y, z, ...]
        sample = np.asarray(sample, dtype = float, order = "C")
        if sample.ndim != 2 or sample.shape[1] < 4:
            raise ValueError((
                "\n[ERROR]: `sample` should have two dimensions (M, N), where "
                f"N >= 4. Received {sample.shape}.\n"
            ))

        labels = self.clusterer.fit_predict(sample[:, 1:4])
        max_label = labels.max()

        # If `allow_single_cluster` is "auto", check if no clusters were found
        # and try again using the hdbscan option allow_single_cluster = True.
        if max_label == -1 and self._allow_single_cluster == "auto":
            labels = self.clusterer_single.fit_predict(sample[:, 1:4])
            max_label = labels.max()

            if self._select_exemplars:
                prediction_data = \
                    self.clusterer_single._prediction_data
                selected_clusters = \
                    self.clusterer_single.condensed_tree_._select_clusters()
                raw_condensed_tree = \
                    self.clusterer_single.condensed_tree_._raw_tree

        elif self._select_exemplars:
            prediction_data = \
                self.clusterer._prediction_data
            selected_clusters = \
                self.clusterer.condensed_tree_._select_clusters()
            raw_condensed_tree = \
                self.clusterer.condensed_tree_._raw_tree

        # Select the cluster exemplars' indices
        if self._select_exemplars:
            exemplars = []
            for cluster in selected_clusters:
                cluster_exemplars = []
                for leaf in prediction_data._recurse_leaf_dfs(cluster):
                    leaf_max_lambda = raw_condensed_tree['lambda_val'][
                        raw_condensed_tree['parent'] == leaf
                    ].max()
                    points = raw_condensed_tree['child'][
                        (raw_condensed_tree['parent'] == leaf) &
                        (raw_condensed_tree['lambda_val'] == leaf_max_lambda)
                    ]
                    cluster_exemplars.append(points)
                exemplars.append(np.hstack(cluster_exemplars))

            exemplars = np.hstack(exemplars)

            # Create a boolean mask for rows in `sample` which are exemplars
            indices_exemplars = np.zeros(len(sample), dtype = bool)
            indices_exemplars[exemplars] = True

        if _set_labels:
            self._labels = labels

        # The tracer location (centre of a cluster's exemplars) is the average
        # of the time, x, y, z columns + the number of points in that cluster
        # (i.e. cluster size)
        # centres columns: [time, x, y, z, ..etc.., cluster_size]
        centres = []
        for i in range(0, max_label + 1):
            # Average time, x, y, z of the exemplars in cluster with label i
            indices_cluster = (labels == i)

            if self._select_exemplars:
                centre = np.mean(sample[indices_cluster & indices_exemplars],
                                 axis = 0)
            else:
                centre = np.mean(sample[indices_cluster], axis = 0)

            # Append the number of points of label i => cluster_size
            centre = np.append(centre, indices_cluster.sum())
            centres.append(centre)

        centres = np.array(centres)

        if not as_array and len(centres) != 0:
            centres = pept.PointData(
                centres,
                sample_size = 0,
                overlap = 0,
                verbose = False
            )

        if verbose:
            end = time.time()
            print("Fitting one sample took {} seconds".format(end - start))

        # If labels are requested, also return the initial sample with appended
        # labels. Labels go from 0 to max_label; -1 represents noise.
        if get_labels:
            sample_labelled = np.append(
                sample, labels[:, np.newaxis], axis = 1
            )

            if not as_array and len(sample_labelled) != 0:
                sample_labelled = pept.PointData(
                    sample_labelled,
                    sample_size = 0,
                    overlap = 0,
                    verbose = False
                )
            return centres, sample_labelled

        # Otherwise just return the found centres
        return centres


    def fit(
        self,
        cutpoints,
        get_labels = False,
        max_workers = None,
        verbose = True
    ):
        '''Fit cutpoints (an instance of `PointData`) and return the cluster
        centres and (optionally) the labelled cutpoints.

        This is a convenience function that clusters each sample in an instance
        of `pept.PointData` *in parallel*, using joblib. For more fine-grained
        control over the clustering, the `fit_sample` method can be used for
        each individual sample.

        Parameters
        ----------
        cutpoints : an instance of `pept.PointData`
            The samples of points that will be clustered. Be careful to set the
            appropriate `sample_size` and `overlap` for good results. If the
            `sample_size` is too low, the less radioactive tracers might not be
            found; if it is too high, temporal resolution is decreased. If the
            `overlap` is too small, the tracked points might be very "sparse".
            Note: when transforming LoRs into cutpoints using the `Cutpoints`
            class, the `sample_size` is automatically set based on the average
            number of cutpoints found per sample of LoRs.

        get_labels : bool, default False
            If set to True, the labelled cutpoints are returned along with the
            centres of the clusters. The labelled cutpoints are a list of
            `pept.PointData` for each sample of cutpoints, with an appended
            column representing the cluster labels (starting from 0; noise is
            encoded as -1).

        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            clustering the samples in `cutpoints`. If unset (`None`), the
            number of threads available on the machine (as returned by
            `os.cpu_count()`) will be used.

        verbose : bool, default True
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        Returns
        -------
        centres : pept.PointData
            The centroids of every cluster found with columns
            `[time, x, y, z, ..., cluster_size]`. They are computed as the
            column-wise average of the points included in each cluster (i.e.
            for each label) or the cluster exemplars (if the `select_exemplars`
            attribute is `True`). Another column is added to the initial data
            in `sample`, signifying the cluster size - that is, the number of
            points included in the cluster.

        labelled_cutpoints : optional, pept.PointData
            Returned if `get_labels` is `True`. It is a `pept.PointData`
            instance in which every sample is the corresponding sample in
            `cutpoints`, but with an appended column representing the label of
            the cluster that the point was associated with. The labels are
            integers starting from 0. The points classified as noise have the
            number -1 associated. Note that the labels are only consistent
            within the same sample; that is, for tracers A and B, if in one
            sample A gets the label 0 and B the label 1, in another sample
            their order might be inversed. The `trajectory_separation` module
            might be used to separate them out.

        Raises
        ------
        TypeError
            If `cutpoints` is not an instance (or a subclass) of
            `pept.PointData`.

        Notes
        -----
        If no clusters were found (i.e. all labels are -1), the returned values
        are empty numpy arrays.
        '''

        if verbose:
            start = time.time()

        if not isinstance(cutpoints, pept.PointData):
            raise TypeError((
                "\n[ERROR]: cutpoints should be an instance of "
                "`pept.PointData` (or any class inheriting from it). Received "
                f"{type(cutpoints)}.\n"
            ))

        # Users might forget to set the sample_size, leaving it to the default
        # value of 0; in that case, all points are returned as a single sample;
        # that might not be the intended behaviour.
        if cutpoints.sample_size == 0:
            warnings.warn(
                textwrap.fill((
                    "\n[WARNING]: The `cutpoints.sample_size` was left to the "
                    "default value of 0, in which case all points are returned"
                    " as a single sample. For a very large number of points, "
                    "this might result in a long function execution time.\n"
                ), replace_whitespace = False),
                RuntimeWarning
            )

        get_labels = bool(get_labels)

        # Fit all samples in `cutpoints` in parallel using joblib
        # Collect all outputs as a list. If verbose, show progress bar with
        # tqdm
        if verbose:
            cutpoints = tqdm(cutpoints)

        if max_workers is None:
            max_workers = os.cpu_count()

        data_list = Parallel(n_jobs = max_workers)(
            delayed(self.fit_sample)(
                sample,
                get_labels = get_labels,
                as_array = True,
                verbose = False,
                _set_labels = False
            ) for sample in cutpoints
        )

        if not get_labels:
            # data_list is a list of arrays. Only choose the arrays with at
            # least one row.
            centres = [r for r in data_list if len(r) != 0]
        else:
            # data_list is a list of tuples, in which the first element is an
            # array of the centres, and the second element is an array of the
            # labelled cutpoints.
            centres = [r[0] for r in data_list if len(r[0]) != 0]

        if len(centres) != 0:
            centres = pept.PointData(
                np.vstack(centres),
                sample_size = 0,
                overlap = 0,
                verbose = False
            )

        if verbose:
            end = time.time()
            print("\nFitting cutpoints took {} seconds.\n".format(end - start))

        if get_labels:
            # data_list is a list of tuples, in which the first element is an
            # array of the centres, and the second element is an array of the
            # labelled cutpoints.
            labelled_cutpoints = [r[1] for r in data_list if len(r[1]) != 0]
            if len(labelled_cutpoints) != 0:
                # Encapsulate `labelled_cutpoints` in a `pept.PointData`
                # instance in which every sample is the corresponding sample in
                # `cutpoints`, but with an appended column representing the
                # labels. Therefore, the `sample_size` is the same as for
                # `cutpoints`, which is equal to the length of every array in
                # `labelled_cutpoints`
                labelled_cutpoints = pept.PointData(
                    np.vstack(labelled_cutpoints),
                    sample_size = len(labelled_cutpoints[0]),
                    overlap = 0,
                    verbose = False
                )

            # Set the attribute `labels` to the stacked labels of all the
            # labelled cutpoints; that is, the last column in the
            # labelled_cutpoints internal data:
            self._labels = labelled_cutpoints.points[:, -1]

            return centres, labelled_cutpoints

        return centres


    def optimise(
        self,
        points,
        nsamples = 16,
        selected_indices = None,
        max_workers = None,
        verbose = True,
        _stability_params = None
    ):
        '''Optimise HDBSCANClusterer settings against a given dataset of
        `points`.

        The `points` are either a single sample (i.e. an (M, N>=4) numpy array)
        or an instance of `PointData` (i.e. multiple samples). In the latter
        case, the settings are optimised against `nsamples` random ramples
        (if `selected_indices` is None), or the samples at indices
        `selected_incides`.

        The CMA-ES (Covariance Matrix Adaptation Evolution Strategy) [1]_
        algorithm is used to find the `min_cluster_size` and
        `min_samples` combination that maximises the overall cluster
        stabilities, quantified with a fitness function that uses:

        1. The probability of each clustered point of being in the assigned
           cluster.
        2. The persistence of each cluster, gauging the relative coherence of
           the clusters.

        As the samples used for optimisation are chosen randomly and CMA-ES is
        a stochastic algorithm, some variance in the results of different
        optimisation passes is expected.

        Parameters
        ----------
        points : (M, N>=4) numpy.ndarray or pept.PointData
            The dataset for which the HDBSCAN parameters will be optimised. It
            is either a single sample (a numpy array) or an instance of
            `PointData` (multiple samples).

        nsamples : int, default 16
            If `points` is an instance of `pept.PointData` and
            `selected_indices` is `None`, `nsamples` random samples are
            selected to optimise against. A larger number may yield better
            parameters, but take more time to find. The default of 16 provides
            a good trade-off between the two.

        selected_indices : list-like of int, optional
            If `None`, `nsamples` random samples are selected from `points`.
            Otherwise, select the samples in `points` at the indices contained
            in `selected_indices`. Only has an effect if `points` is a
            `pept.PointData`.

        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            clustering the selected samples in `points`. If unset (`None`), the
            number of threads available on the machine (as returned by
            `os.cpu_count()`) will be used.

        verbose : bool, default True
            Provide extra information during the optimisation pass.

        _stability_params : list-like of length 2 or 4, optional
            If set, the function will no longer run optimisations; rather, it
            will only return the negative fitness score, computed for the
            parameters in `_stability_params`. If `allow_single_cluster` is
            "auto", this parameter should contain the `min_cluster_size` and
            `min_samples` values for `self.clusterer`, followed by the values
            for `self.clusterer_single`. Otherwise, it should contain only the
            values for `self.clusterer`. This is useful for debugging and
            using a different algorithm than CMA-ES.

        Returns
        -------
        stabilities : float or list[float] of length 2
            If `_stability_params` is not `None`, this method will return the
            cluster stability scores. If `allow_single_cluster` is "auto", it
            will return the scores for both the `clusterer` and
            `clusterer_single`; otherwise it returns only the `clusterer`
            score. If `_stability_params` is `None`, nothing is returned.

        Raises
        ------
        ValueError
            If `selected_indices` is not `None` and is not a list-like with a
            single dimension.

        ValueError
            If `points` is a single sample and does not have shape (M, N>=4).

        Notes
        -----
        If the optimisation is successful, this class' `min_cluster_size` and
        `min_samples` attributes will be automatically set to the optimum
        values found.

        For datasets with 2000 points per sample, the optimisation should have
        around 200 function evaluations (i.e. clustering runs) and take
        somewhere under 2 minutes.

        References
        ----------
        .. [1] Nikolaus Hansen, Youhei Akimoto, and Petr Baudis. CMA-ES/pycma
           on Github. Zenodo, DOI:10.5281/zenodo.2559634, February 2019.

        '''

        # Select samples for clustering
        # points is either a single sample or PointData (i.e. multiple samples)
        if isinstance(points, pept.PointData):
            if selected_indices is None:
                # Select `nsamples` random samples
                rng = np.random.default_rng()
                selected_indices = rng.integers(0, len(points) - 1, nsamples)
            else:
                # Ensure `selected_indices` is a list-like of integers
                selected_indices = np.asarray(selected_indices, dtype = int)
                if selected_indices.ndim != 1:
                    raise ValueError(textwrap.fill(
                        "[ERROR]: If `selected_indices` is defined, it must "
                        "be list-like with a single dimension. Received "
                        f"{selected_indices.shape}"
                    ))

            selected = [points[i] for i in selected_indices]

        else:
            # We only have a single sample
            points = np.asarray(points, dtype = float)
            if points.ndim != 2 or points.shape[1] < 4:
                raise ValueError(textwrap.fill(
                    "[ERROR]: If `points` is a single sample, it should have "
                    f"shape (M, N>=4). Received {points.shape}."
                ))

            selected = [points]

        # If `_stability_params` were given, don't optimise and only return the
        # stability scores.
        if _stability_params is not None:
            score = self._optimise(
                self.clusterer, selected, max_workers, verbose,
                _stability_params[:2]
            )

            if self.clusterer_single is not None:
                self.clusterer.min_cluster_size = \
                    int(round(_stability_params[0]))
                self.clusterer.min_samples = int(round(_stability_params[1]))

                score_single = self._optimise_single(
                    self.clusterer_single, self.clusterer, selected,
                    max_workers, verbose, _stability_params[2:]
                )

                return score, score_single

            return score

        # Print optimisation info to the terminal.
        if verbose:
            print((
                "\n" + "-" * 79 + "\n" +
                "Optimising HDBSCANClusterer settings against given "
                "dataset.\n" + "-" * 79 + "\n"
            ))

            if self.clusterer_single is not None:
                print((
                    '`allow_single_cluster` is "auto" - will optimise '
                    "clusterers individually.\n"
                ))

        # Optimise `self.clusterer` settings. `self.clusterer_single` is not
        # used here!
        min_cluster_size, min_samples, score = self._optimise(
            self.clusterer, selected, max_workers, verbose, _stability_params
        )

        # Warning if no feasible settings / clusters were found. Otherwise set
        # min_cluster_size and min_samples to the found optimal values.
        if score == 0.0:
            warnings.warn(
                textwrap.fill((
                    "\n[WARNING]: Could not find settings combination that "
                    "improves the clusters' stability. Are there any clusters "
                    "in the input dataset?\nDid not set `min_cluster_size` "
                    "and `min_samples`.\n"
                ), replace_whitespace = False),
                RuntimeWarning
            )
        else:
            self.min_cluster_size = min_cluster_size
            self.min_samples = min_samples

        # If `allow_single_cluster` == "auto" (i.e. clusterer_single != None),
        # also optimise `self.clusterer_sigle` settings, only for the samples
        # in which `self.clusterer` does not find anything.
        if self.clusterer_single is not None:
            min_cluster_size_single, min_samples_single, score_single = \
                self._optimise_single(
                    self.clusterer_single, self.clusterer, selected,
                    max_workers, verbose, _stability_params
                )

            # Warning if no samples with single clusters were found. Otherwise
            # set `min_cluster_size_single` and `min_samples` to the found
            # optimal values.
            if score_single == 0.0:
                warnings.warn(
                    textwrap.fill((
                        "\n[WARNING]: Could not find settings combination that"
                        " improves the clusters' stability for the single "
                        "cluster case. Are there any samples with a single "
                        "tracer in the input dataset?\n\n"
                        "Setting `min_cluster_size` and `min_samples` to the "
                        "multi-tracer settings.\n"
                    ), replace_whitespace = False),
                    RuntimeWarning
                )
                self.min_cluster_size_single = min_cluster_size
                self.min_samples_single = min_samples
            else:
                self.min_cluster_size_single = min_cluster_size_single
                self.min_samples_single = min_samples_single

        # Print optimisation results to terminal.
        if verbose:
            print((
                "-" * 79 +
                "\nOptimal parameters found:\n"
                f"min_cluster_size = {self.min_cluster_size}\n"
                f"min_samples =      {self.min_samples}"
            ))

            if self.clusterer_single is not None:
                print((
                    "\n"
                    f"min_cluster_size_single = {self.min_cluster_size_single}"
                    f"\nmin_samples_single =      {self.min_samples_single}"
                ))

            print("-" * 79 + "\n\n")


    def _optimise(
        self,
        clusterer,
        selected,
        max_workers = None,
        verbose = True,
        _stability_params = None
    ):
        if verbose:
            start = time.time()

        if max_workers is None:
            max_workers = os.cpu_count()

        def stability_sample(clusterer, sample):
            # Only cluster based on [x, y, z].
            labels = clusterer.fit_predict(sample[:, 1:4])
            max_label = labels.max()

            prob = clusterer.probabilities_
            pers = clusterer.cluster_persistence_

            score = 0.0
            for i, label in enumerate(range(0, max_label + 1)):
                score += prob[labels == label].mean() * pers[i]

            return -score / (max_label + 2)

        def stability(params):
            params = np.asarray(params)
            clusterer.min_cluster_size = int(round(params[0]))
            clusterer.min_samples = int(round(params[1]))

            # Compute stabilities for each selected sample in parallel
            stabilities = Parallel(n_jobs = max_workers)(
                delayed(stability_sample)(clusterer, sample)
                for sample in selected
            )

            stabilities = np.array(stabilities)
            score = stabilities.mean()

            # Decrease score if there is more than 50% difference between
            # `min_cluster_size` and `min_samples`
            correction = params.min() / params.max()

            return score * correction if correction < 0.5 else score

        if _stability_params is not None:
            return stability(_stability_params)

        sample_size = len(selected[0])
        x0 = [2, 2]                 # starting guess = smallest parameters
        sigma0 = 0.1 * sample_size  # initial standard deviation
        bounds = [2, sample_size]   # bounds for min_{cluster_size | samples}

        es = cma.CMAEvolutionStrategy(x0, sigma0, dict(
            bounds = bounds,
            integer_variables = [0, 1],
            tolflatfitness = 10,                # Try 10 times if f is flat
            tolfun = -np.inf,                   # Only stop when tolx < 1
            tolx = 1,                           # Stop if changes in x < 1
            verbose = 3 if verbose else -9
        ))
        es.optimize(stability)

        min_cluster_size = int(round(es.result.xfavorite[0]))
        min_samples = int(round(es.result.xfavorite[1]))
        score = es.result.fbest

        if verbose:
            es.result_pretty()

            end = time.time()
            print(f"\nOptimisation took {end - start} s.\n")

        return min_cluster_size, min_samples, score


    def _optimise_single(
        self,
        clusterer_single,
        clusterer_multi,
        selected,
        max_workers = None,
        verbose = True,
        _stability_params = None
    ):

        if verbose:
            start = time.time()

        if max_workers is None:
            max_workers = os.cpu_count()

        def stability_sample(clusterer_single, clusterer_multi, sample):
            # Only cluster based on [x, y, z].
            labels = clusterer_multi.fit_predict(sample[:, 1:4])
            max_label = labels.max()

            # Only score if single cluster...
            if max_label != -1:
                return 0

            labels = clusterer_single.fit_predict(sample[:, 1:4])
            max_label = labels.max()

            prob = clusterer_single.probabilities_
            pers = clusterer_single.cluster_persistence_

            score = 0.0
            for i, label in enumerate(range(0, max_label + 1)):
                score += prob[labels == label].mean() * pers[i]

            return -score / (max_label + 2)

        def stability(params):
            params = np.asarray(params)
            clusterer_single.min_cluster_size = int(round(params[0]))
            clusterer_single.min_samples = int(round(params[1]))

            # Compute stabilities for each selected sample in parallel
            stabilities = Parallel(n_jobs = max_workers)(
                delayed(stability_sample)(
                    clusterer_single, clusterer_multi, sample
                ) for sample in selected
            )

            stabilities = np.array(stabilities)
            score = stabilities.mean()

            # Decrease score if there is more than 50% difference between
            # `min_cluster_size` and `min_samples`
            correction = params.min() / params.max()

            return score * correction if correction < 0.5 else score

        if _stability_params is not None:
            return stability(_stability_params)

        sample_size = len(selected[0])
        x0 = [2, 2]                 # starting guess = smallest parameters
        sigma0 = 0.1 * sample_size  # initial standard deviation
        bounds = [2, sample_size]   # bounds for min_{cluster_size | samples}

        es = cma.CMAEvolutionStrategy(x0, sigma0, dict(
            bounds = bounds,
            integer_variables = [0, 1],
            tolflatfitness = 10,                # Try 10 times if f is flat
            tolfun = -np.inf,                   # Only stop when tolx < 1
            tolx = 1,                           # Stop if changes in x < 1
            verbose = 3 if verbose else -9
        ))
        es.optimize(stability)

        min_cluster_size = int(round(es.result.xfavorite[0]))
        min_samples = int(round(es.result.xfavorite[1]))
        score = es.result.fbest

        if verbose:
            es.result_pretty()

            end = time.time()
            print(f"\nOptimisation took {end - start} s.\n")

        return min_cluster_size, min_samples, score


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"clusterer:\n{self.clusterer}\n\n"
            f"min_cluster_size =            {self.min_cluster_size}\n"
            f"min_samples =                 {self.min_samples}\n\n"
            f"clusterer_single:\n{self.clusterer_single}\n\n"
            f"min_cluster_size_single =     {self.min_cluster_size_single}\n"
            f"min_samples_single =          {self.min_samples_single}\n\n"
            f"allow_single_cluster =        {self.allow_single_cluster}\n"
            f"labels =                      {self.labels}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL
        docstr = (
            "Class instance that inherits from `peptml.HDBSCANClusterer`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n"
            "----------\n"
            f"{self.__str__()}\n"
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 14.01.2020


from    .traverse2d     import  traverse2d
from    .traverse3d     import  traverse3d

__all__ = [
    "traverse2d",
    "traverse3d",
]


__license__ =       "GNU v3.0"
__maintainer__ =    "Andrei Leonard Nicusan"
__email__ =         "a.l.nicusan@bham.ac.uk"
__status__ =        "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : PEPTsimulator.py
# License           : License: GNU v3.0
# Author            : Andrei Leonard Nicusan <aln705@student.bham.ac.uk>
# Date              : 01.07.2019


import  numpy                   as          np

from    tqdm                    import      tqdm




class Shape:
    '''Class of shape functions which return a random [x, y, z] point inside a
    given shape (sphere, cylinder, etc) centred around the origin.
    '''

    def __init__(self, x=0.5, y=0.5, z=0.5):
        '''x, y, z are coordinate ranges for the three
        major axes. For example:
        For a sphere, x is radius; y and z are unused
        For a cylinder, x is radius, y is height; z is unused
        For a parallelipiped, x is width, y is depth, z is height

        '''
        self.x = x
        self.y = y
        self.z = z


    def rotateX3D(self, vec, angleX):
        '''Rotate vec around the X axis by angleX radians.
        '''
        rot = np.array([
            [1, 0, 0],
            [0, np.cos(angleX), -np.sin(angleX)],
            [0, np.sin(angleX), np.cos(angleX)],
        ])

        return (rot @ vec)



    def sphere(self):
        '''Simulate as spherical coordinates.
        '''
        r = np.random.uniform(0, self.x)
        theta = np.random.uniform(0, np.pi)
        phi = np.random.uniform(0, 2 * np.pi)

        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)

        return [x, y, z]


    def cylinder(self, angleX=0, angleY=0, angleZ=0):
        '''Simulate as cylindrical coordinates.

        Cylinder is horizontal on the x axis
        => self.x is max radius
        => self.y is max height
        '''
        r = np.random.uniform(0, self.x)

        theta = np.random.uniform(0, 2 * np.pi)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        z = np.random.uniform(-self.y, self.y)

        return [x, y, z]




class Noise:
    '''Add noise to simulated PEPT data.
    '''

    def __init__(self, trajectory, xMax, yMax, zMax):
        '''Trajectory row: [time, X, Y, Z]
        '''
        self.trajectory = trajectory
        self.xMax = xMax
        self.yMax = yMax
        self.zMax = zMax


    # Better to use addNoiseToPEPT. Add noise to PEPT screen projection, rather
    # than to the actual trajectory


    def create_trajectory_noise(self, numberOfPoints):
        '''Generate random X, Y, Z coordinates for noise points

        '''
        xNoise = np.random.uniform(0, self.xMax, numberOfPoints)
        yNoise = np.random.uniform(0, self.yMax, numberOfPoints)
        zNoise = np.random.uniform(0, self.zMax, numberOfPoints)

        # Generate times for noise points between the start time and end time
        # of given trajectory
        timeNoise = np.random.uniform(
            self.trajectory[0][0],
            self.trajectory[-1][0],
            numberOfPoints,
        )
        timeNoise = np.sort(timeNoise)

        self.noise = np.stack((timeNoise, xNoise, yNoise, zNoise), axis=1)
        return self.noise


    def get_trajectory_with_noise(self, noiseRatio):
        '''Get trajectory with noise.
        '''

        self.numberOfNoisePoints = int(len(self.trajectory) * noiseRatio)
        self.createTrajectoryNoise(self.numberOfNoisePoints)

        # Find indices at which the noise should be inserted to maintain time
        # order of trajectory
        insertionIndices = np.searchsorted(
            self.trajectory[:, 0],
            self.noise[:, 0],
            side='left'
        )

        self.trajectoryWithNoise = np.insert(
            self.trajectory,
            insertionIndices,
            self.noise,
            axis=0,
        )

        return self.trajectoryWithNoise


    def add_noise_to_pept(self, pept_data, number_of_noise_points):
        '''PEPTdata row: [time, X1, Y1, X2, Y2]
        Generate random (X, Y) pairs coordinates for noise points in PEPTdata
        '''

        x1Noise = np.random.uniform(0, self.xMax, number_of_noise_points)
        y1Noise = np.random.uniform(0, self.yMax, number_of_noise_points)

        x2Noise = np.random.uniform(0, self.xMax, number_of_noise_points)
        y2Noise = np.random.uniform(0, self.yMax, number_of_noise_points)

        # Generate times for noise lines between the start time and end time of
        # given trajectory
        time_noise = np.random.uniform(
            self.trajectory[0][0],
            self.trajectory[-1][0],
            number_of_noise_points
        )
        time_noise = np.sort(time_noise)

        self.pept_noise = np.stack(
            (time_noise, x1Noise, y1Noise, x2Noise, y2Noise),
            axis=1,
        )
        insertion_indices = np.searchsorted(
            pept_data[:, 0],
            self.pept_noise[:, 0],
            side='left',
        )
        self.pept_data_with_noise = np.insert(
            pept_data,
            insertion_indices,
            self.pept_noise,
            axis=0,
        )

        return self.pept_data_with_noise


    def add_spread_to_pept(self, pept_data, max_spread, depth):
        '''PEPTdata row: [time, X1, Y1, X2, Y2]
        '''

        pept_data_copy = np.copy(pept_data)

        # Add positional spread to X1, Y1, X2, Y2
        spread = np.random.uniform(-max_spread, max_spread,
                                   (len(pept_data), 4))
        pept_data_copy[:, 1:5] = pept_data_copy[:, 1:5] + spread

        # Add angular spread to X1, Y1, X2, Y2
        # x1_angular_spread = x1 - (x2 - x1) / sep * depth * alpha
        alpha = np.random.uniform(0, 1, (len(pept_data), 4))

        self.pept_data_with_spread = np.copy(pept_data_copy)
        self.pept_data_with_spread[:, 1] = pept_data_copy[:, 1] - \
            (pept_data_copy[:, 3] - pept_data_copy[:, 1]) / \
            self.zMax * alpha[:, 0] * depth

        self.pept_data_with_spread[:, 3] = pept_data_copy[:, 3] + \
            (pept_data_copy[:, 3] - pept_data_copy[:, 1]) / \
            self.zMax * alpha[:, 1] * depth

        self.pept_data_with_spread[:, 2] = pept_data_copy[:, 2] - \
            (pept_data_copy[:, 4] - pept_data_copy[:, 2]) / \
            self.zMax * alpha[:, 2] * depth
        self.pept_data_with_spread[:, 4] = pept_data_copy[:, 4] + \
            (pept_data_copy[:, 4] - pept_data_copy[:, 2]) / \
            self.zMax * alpha[:, 3] * depth

        return self.pept_data_with_spread




class Simulator:
    '''Simulate PEPT data.
    '''

    def __init__(
        self,
        trajectory,
        sampling_times,
        shape_function,
        separation = 712,
        decay_energy = 0.6335,
        Zeff = 7.22,
        Aeff = 13,
        x_max = 500,
        y_max = 500
    ):
        '''Simulator class constructor.
        '''

        # Trajectory row: [time, X, Y, Z]
        self.trajectory = trajectory

        # Timepoints at which positrons are emitted
        self.sampling_times = sampling_times

        # Separation between PEPT screens
        self.separation = separation

        # Function which returns an [x, y, z] positions around a centre
        self.shape_function = shape_function

        # Beta+ endpoint/decay energy of tracer, in MeV. F-18 has a Beta+ decay
        # energy of 0.6335 MeV
        self.decay_energy = decay_energy

        # Effective atomic weight Aeff and atomic number Zeff. For water,
        # Aeff = 13, Zeff = 7.22
        self.Aeff = Aeff
        self.Zeff = Zeff

        # PEPT screens sizes
        self.x_max = x_max
        self.y_max = y_max


    def simulate(self):

        # Indices of trajectory times closest to sampling times
        location_indices = np.searchsorted(
            self.trajectory[:, 0],
            self.sampling_times,
            side='left',
        )
        number_of_samples = len(location_indices)

        # PEPTdata row: [time, X1, Y1, X2, Y2]
        pept_data = []

        # For every position on the trajectory corresponding to a samplingTime,
        # generate a PEPTdata row
        for i in tqdm(range(0, number_of_samples)):

            # particle index in the trajectory for ray tracing
            particle_index = location_indices[i]

            # particleIndex for insertion could be the last one
            if particle_index >= len(self.trajectory):
                particle_index = len(self.trajectory) - 1

            # The positron emission point within the particle
            [xshape, yshape, zshape] = self.shape_function()
            xp = self.trajectory[particle_index][1] + xshape
            yp = self.trajectory[particle_index][2] + yshape
            zp = self.trajectory[particle_index][3] + zshape

            # Calculate incident/kinetic energy Ei of positron from
            # distribution approximated as gaussian

            # mean
            mu = self.decay_energy / 2
            # 99.7% of data will lie between [0, decayEnergy]
            sigma = mu / 3

            Ei = np.random.normal(mu, sigma)

            # Find positron annihilation point, corresponding to gamma
            # radiation emission using Palmer et al., 1992 equations
            b1 = 4.569 * self.Aeff / self.Zeff**1.209
            b2 = 1 / (2.873 - 0.02309 * self.Zeff)
            Rex = b1 * Ei * Ei / (b2 + Ei)

            # Annihilation point [xa, ya, za] from positron emission point
            [xa, ya, za] = np.random.normal(0, Rex / 2, 3)

            # The gamma ray emission point will therefore be
            # positron emission point + positron annihilation point
            xp += xa
            yp += ya
            zp += za

            # Try at most 100 random phi and theta angles until the ray falls
            # within the PEPT screens
            ray_try = 0
            while ray_try < 100:
                phi = np.random.uniform(0, np.pi)
                theta = np.random.uniform(0, np.pi)

                if phi == np.pi / 2 or theta == np.pi / 2:
                    continue

                x1 = xp - zp * np.tan(phi)
                x2 = xp + (self.separation - zp) * np.tan(phi)

                y1 = yp - zp / np.cos(phi) * np.tan(theta)
                y2 = yp + (self.separation - zp) / np.cos(phi) * np.tan(theta)

                # Check the rays fell within the PEPT screens
                if 0 < x1 < self.x_max and 0 < x2 < self.x_max:
                    if 0 < y1 < self.y_max and 0 < y2 < self.y_max:
                        pept_data.append(np.array([
                            self.sampling_times[i], x1, y1, x2, y2
                        ]))
                        break
                ray_try += 1

        self.pept_data = np.array(pept_data)
        return self.pept_data


    def add_noise(self, noise_ratio):
        noiser = Noise(
            self.trajectory, self.x_max, self.y_max, self.separation
        )
        number_of_noise_points = int(noise_ratio * len(self.sampling_times))
        self.pept_data_with_noise = noiser.add_noise_to_pept(
            self.pept_data, number_of_noise_points
        )


    def add_spread(self, max_spread = 4, depth = 16):
        noiser = Noise(
            self.trajectory, self.x_max, self.y_max, self.separation
        )
        self.pept_data_with_noise = noiser.add_spread_to_pept(
            self.pept_data, max_spread, depth
        )


    def add_noise_and_spread(self, noise_ratio, max_spread = 4, depth = 16):
        noiser = Noise(
            self.trajectory, self.x_max, self.y_max, self.separation
        )
        number_of_noise_points = int(noise_ratio * len(self.sampling_times))
        self.pept_data_with_noise = noiser.add_noise_to_pept(
            self.pept_data, number_of_noise_points
        )
        self.pept_data_with_noise = noiser.add_spread_to_pept(
            self.pept_data_with_noise, max_spread, depth
        )


    def change_trajectory(self, new_trajectory):
        self.trajectory = new_trajectory


    def change_sampling_times(self, new_sampling_times):
        self.sampling_times = new_sampling_times


    def change_shape(self, new_shape_function):
        self.shape_function = new_shape_function


    def write_csv(self, fname):
        np.savetxt(fname, self.pept_data, delimiter = '   ', newline = '\n  ')


    def write_noise_csv(self, fname):
        np.savetxt(fname, self.pept_data_with_noise, delimiter = '   ',
                   newline = '\n  ')



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : occupancy_ext.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 23.11.2020


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


#import numpy as np

from libc.math cimport floor, sqrt, acos, M_PI
from libc.float cimport DBL_MAX


cdef inline double fabs(double x) nogil:
    return (x if x >= 0 else -x)


cdef inline double dist(double x1, double y1, double x2, double y2) nogil:
    # Distance between two 2D points
    return sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)


cdef inline double dist2(double x1, double y1, double x2, double y2) nogil:
    # Squared distance between two 2D points
    return (x1 - x2) ** 2 + (y1 - y2) ** 2


cdef inline double term(double r, double _c, double _p) nogil:
    # Term in calculation of `one_corner`
    # Same for xc / xp and yc / yp
    return sqrt(r ** 2 - (_c - _p) ** 2)


cdef inline double triangle(
    double x1, double y1, double x2, double y2, double x3, double y3
) nogil:
    return 0.5 * fabs(x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2))


cdef inline double dist_pl(
    double x0,
    double y0,
    double x1,
    double y1,
    double x2,
    double y2
) nogil:
    # Distance from a point (x0, y0) to a line defined by two points
    # (x1, y1) and (x2, y2)
    denom = dist(x1, y1, x2, y2)
    nom = fabs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)

    return nom / denom


cdef inline double circular_segment(double r, double h) nogil:
    # Area of a circular segment from a circle of radius r and a line
    # If the distance from the circle centre to the line is d, then
    # h = r - d
    if h < 0.0:     # Happens for very small distances (d = r * 1e-13)
        return 0.0

    return r ** 2 * acos((r - h) / r) - (r - h) * sqrt(2 * r * h - h ** 2)


cdef inline double zero_corner(
    Py_ssize_t[4] corners,
    double px,
    double py,
    double xsize,
    double ysize,
    double x0,
    double y0,
    double r,
) nogil:
    '''Compute the overlapping area between a circle and a rectangle when
    exactly one of the rectangle's corners are inside the circle.

    Parameters
    ----------
    corners
        Vector of 0 and 1 for the corner that is inside the circle. The
        corners are in trigonometric order.

    px, py
        The pixel x and y coordinates.

    xsize, ysize
        The pixel width and height.

    x, y
        The circle origin x and y coordinates

    r
        The circle radius

    Returns
    -------
    The overlapping area.
    '''

    # Distance from circle centre x0, y0 to line defined by the two
    # intersections above and h = (r - d)
    cdef double h
    cdef double area = 0.
    cdef double xmin = px - xsize / 2
    cdef double xmax = px + xsize / 2
    cdef double ymin = py - ysize / 2
    cdef double ymax = py + ysize / 2

    # Corner case - if zero corners are inside, there might be no intersection
    # This happens when the centre of the particle is in one of the corners
    if ((x0 <= xmin and y0 <= ymin) or      # Bottom left
            (x0 >= xmax and y0 <= ymin) or  # Bottom right
            (x0 >= xmax and y0 >= ymax) or  # Top right
            (x0 <= xmin and y0 >= ymax)):   # Top left
        return 0.

    # Particle is fully encapsulated
    if xmin < x0 < xmax and ymin < y0 < ymax:
        return M_PI * r * r
    # Below
    elif xmin < x0 < xmax:
        h = ysize / 2 + r - fabs(py - y0)
    # Above
    else:
        h = xsize / 2 + r - fabs(px - x0)

    # Area of the circular segment
    area += circular_segment(r, h)

    return area


cdef inline double one_corner(
    Py_ssize_t[4] corners,
    double px,
    double py,
    double xsize,
    double ysize,
    double x0,
    double y0,
    double r,
) nogil:
    '''Compute the overlapping area between a circle and a rectangle when
    exactly one of the rectangle's corners are inside the circle.

    Parameters
    ----------
    corners
        Vector of 0 and 1 for the corner that is inside the circle. The
        corners are in trigonometric order.

    px, py
        The pixel x and y coordinates.

    xsize, ysize
        The pixel width and height.

    x, y
        The circle origin x and y coordinates

    r
        The circle radius

    Returns
    -------
    The overlapping area.
    '''

    # The bounded corner coordinates
    cdef double xc = 0, yc = 0

    # The coords of the two intersections between the circle and the rectangle
    cdef double x1 = 0, y1 = 0, x2 = 0, y2 = 0

    # Distance from circle centre x0, y0 to line defined by the two
    # intersections above and h = (r - d)
    cdef double d, h

    cdef double area = 0.

    # Pixel boundaries
    cdef double xmin = px - xsize / 2
    cdef double xmax = px + xsize / 2
    cdef double ymin = py - ysize / 2
    cdef double ymax = py + ysize / 2

    # Lower left corner
    if corners[0] == 1:
        xc = xmin
        yc = ymin

        x1 = x0 + term(r, yc, y0)
        y1 = yc

        x2 = xc
        y2 = y0 + term(r, xc, x0)

    # Lower right corner
    elif corners[1] == 1:
        xc = xmax
        yc = ymin

        x1 = xc
        y1 = y0 + term(r, xc, x0)

        x2 = x0 - term(r, yc, y0)
        y2 = yc

    # Upper right corner
    elif corners[2] == 1:
        xc = xmax
        yc = ymax

        x1 = x0 - term(r, yc, y0)
        y1 = yc

        x2 = xc
        y2 = y0 - term(r, xc, x0)

    # Upper left corner
    elif corners[3] == 1:
        xc = xmin
        yc = ymax

        x1 = x0 + term(r, yc, y0)
        y1 = yc

        x2 = xc
        y2 = y0 - term(r, xc, x0)

    d = dist_pl(x0, y0, x1, y1, x2, y2)
    h = r - d

    # Area of the circular segment
    area += circular_segment(r, h)

    # Area of the triangle
    area += dist(xc, yc, x1, y1) * dist(xc, yc, x2, y2) / 2

    return area


cdef inline double two_corner(
    Py_ssize_t[4] corners,
    double px,
    double py,
    double xsize,
    double ysize,
    double x0,
    double y0,
    double r,
) nogil:
    '''Compute the overlapping area between a circle and a rectangle when
    exactly one of the rectangle's corners are inside the circle.

    Parameters
    ----------
    corners
        Vector of 0 and 1 for the corner that is inside the circle. The
        corners are in trigonometric order.

    px, py
        The pixel x and y coordinates.

    xsize, ysize
        The pixel width and height.

    x, y
        The circle origin x and y coordinates

    r
        The circle radius

    Returns
    -------
    The overlapping area.
    '''

    # The bounded corner coordinates
    cdef double xc1 = 0, yc1 = 0, xc2 = 0, yc2 = 0

    # The coords of the two intersections between the circle and the rectangle
    cdef double x1 = 0, y1 = 0, x2 = 0, y2 = 0

    # Distance from circle centre x0, y0 to line defined by the two
    # intersections above and h = (r - d)
    cdef double d, h

    cdef double area = 0.

    # Pixel boundaries
    cdef double xmin = px - xsize / 2
    cdef double xmax = px + xsize / 2
    cdef double ymin = py - ysize / 2
    cdef double ymax = py + ysize / 2

    # Bottom corners
    if corners[0] == 1 and corners[1] == 1:
        xc1 = xmin
        yc1 = ymin

        xc2 = xmax
        yc2 = ymin

        x1 = xc1
        y1 = y0 + term(r, xc1, x0)

        x2 = xc2
        y2 = y0 + term(r, xc2, x0)

    # Top corners
    elif corners[2] == 1 and corners[3] == 1:
        xc1 = xmin
        yc1 = ymax

        xc2 = xmax
        yc2 = ymax

        x1 = xc1
        y1 = y0 - term(r, xc1, x0)

        x2 = xc2
        y2 = y0 - term(r, xc2, x0)

    # Right corners
    elif corners[1] == 1 and corners[2] == 1:
        xc1 = xmax
        yc1 = ymin

        xc2 = xmax
        yc2 = ymax

        x1 = x0 - term(r, yc1, y0)
        y1 = yc1

        x2 = x0 - term(r, yc2, y0)
        y2 = yc2

    # Left corners
    elif corners[0] == 1 and corners[3] == 1:
        xc1 = xmin
        yc1 = ymin

        xc2 = xmin
        yc2 = ymax

        x1 = x0 + term(r, yc1, y0)
        y1 = yc1

        x2 = x0 + term(r, yc2, y0)
        y2 = yc2

    d = dist_pl(x0, y0, x1, y1, x2, y2)
    h = r - d

    # Area of the circular segment
    area += circular_segment(r, h)

    # Area of triangle1 and triangle2
    area += triangle(x1, y1, xc2, yc2, xc1, yc1)
    area += triangle(x1, y1, xc2, yc2, x2, y2)

    return area


cdef inline double three_corner(
    Py_ssize_t[4] corners,
    double px,
    double py,
    double xsize,
    double ysize,
    double x0,
    double y0,
    double r,
) nogil:
    '''Compute the overlapping area between a circle and a rectangle when
    exactly one of the rectangle's corners are inside the circle.

    Parameters
    ----------
    corners
        Vector of 0 and 1 for the corner that is inside the circle. The
        corners are in trigonometric order.

    px, py
        The pixel x and y coordinates.

    xsize, ysize
        The pixel width and height.

    x, y
        The circle origin x and y coordinates

    r
        The circle radius

    Returns
    -------
    The overlapping area.
    '''

    # The bounded corner coordinates
    cdef double xc1 = 0, yc1 = 0, xc2 = 0, yc2 = 0
    cdef double xc3 = 0, yc3 = 0, xc4 = 0, yc4 = 0

    # The coords of the two intersections between the circle and the rectangle
    cdef double x1 = 0, y1 = 0, x2 = 0, y2 = 0

    # Distance from circle centre x0, y0 to line defined by the two
    # intersections above and h = (r - d)
    cdef double d, h

    cdef double area = 0.

    # Pixel boundaries
    cdef double xmin = px - xsize / 2
    cdef double xmax = px + xsize / 2
    cdef double ymin = py - ysize / 2
    cdef double ymax = py + ysize / 2

    # Bottom right three corners
    if corners[0] == 1 and corners[1] == 1 and corners[2] == 1:
        xc1 = xmin
        yc1 = ymin

        xc2 = xmax
        yc2 = ymin

        xc3 = xmax
        yc3 = ymax

        xc4 = xmin    # Corner outside
        yc4 = ymax    # Corner outside

        x1 = x0 - term(r, yc3, y0)
        y1 = yc3

        x2 = xc1
        y2 = y0 + term(r, xc1, x0)

    # Top right three corners
    elif corners[1] == 1 and corners[2] == 1 and corners[3] == 1:
        xc1 = xmax
        yc1 = ymin

        xc2 = xmax
        yc2 = ymax

        xc3 = xmin
        yc3 = ymax

        xc4 = xmin    # Corner outside
        yc4 = ymin    # Corner outside

        x1 = xc3
        y1 = y0 - term(r, xc3, x0)

        x2 = x0 - term(r, yc1, y0)
        y2 = yc1

    # Top left three corners
    elif corners[2] == 1 and corners[3] == 1 and corners[0] == 1:
        xc1 = xmax
        yc1 = ymax

        xc2 = xmin
        yc2 = ymax

        xc3 = xmin
        yc3 = ymin

        xc4 = xmax    # Corner outside
        yc4 = ymin    # Corner outside

        x1 = xc1
        y1 = y0 - term(r, xc1, x0)

        x2 = x0 + term(r, yc3, y0)
        y2 = yc3

    # Bottom left three corners
    elif corners[0] == 1 and corners[1] == 1 and corners[3] == 1:
        xc1 = xmin
        yc1 = ymax

        xc2 = xmin
        yc2 = ymin

        xc3 = xmax
        yc3 = ymin

        xc4 = xmax    # Corner outside
        yc4 = ymax    # Corner outside

        x1 = xc3
        y1 = y0 + term(r, xc3, x0)

        x2 = x0 + term(r, yc1, y0)
        y2 = yc1

    d = dist_pl(x0, y0, x1, y1, x2, y2)
    h = r - d

    # Area of the circular segment
    area += circular_segment(r, h)

    # Area of the rectangle
    area += xsize * ysize

    # Area of triangle
    area -= dist(x1, y1, xc4, yc4) * dist(x2, y2, xc4, yc4) / 2

    return area


cdef inline double four_corner(
    Py_ssize_t[4] corners,
    double px,
    double py,
    double xsize,
    double ysize,
    double x0,
    double y0,
    double r,
) nogil:
    '''Compute the overlapping area between a circle and a rectangle when
    exactly one of the rectangle's corners are inside the circle.

    Parameters
    ----------
    corners
        Vector of 0 and 1 for the corner that is inside the circle. The
        corners are in trigonometric order.

    px, py
        The pixel x and y coordinates.

    xsize, ysize
        The pixel width and height.

    x, y
        The circle origin x and y coordinates

    r
        The circle radius

    Returns
    -------
    The overlapping area.
    '''

    return xsize * ysize



cpdef void circles2d_ext(
    double[:, :] pixels,
    double[:, :] positions,
    double[:] radii,
    double[:] xlim,
    double[:] ylim,
) nogil:
    '''Compute the 2D occupancy of circles of different radii.

    ::

        Function signature:

            void circles2d_ext(
                double[:, :] pixels,
                double[:, :] positions,
                double[:] radii,
                double[:] xlim,
                double[:] ylim,
            ) nogil

    This corresponds to the pixellisation of circular particles, such that
    each pixel's value corresponds to the area covered by the particle.

    It is important that all parameters sent to this function are initialised
    properly. Please see the parameters below, or the function *will* segfault.
    Alternatively, use the `pept.processing.occupancy2d` function which is
    more robust and initialises all parameters for you.

    Parameters
    ----------
    pixels: (M, N) numpy.ndarray[ndim = 2, dtype = numpy.float64]
        The 2D grid of pixels, initialised to zero. It can be created with
        `numpy.zeros((nrows, ncols))`.

    positions: (P, 2) numpy.ndarray[ndim = 2, dtype = numpy.float64]
        The particles' 2D positions, where all rows are formatted as
        `[x_coordinate, y_coordinate]`.

    radii: (P,) numpy.ndarray[ndim = 1, dtype = numpy.float64]
        The radii of each particle. It must have the same length as
        `positions`.

    xlim: (2,) numpy.ndarray[ndim = 1, dtype = numpy.float64]
        The limits of the system over which the pixels span in the
        x-dimension, formatted as [xmin, xmax].

    ylim: (2,) numpy.ndarray[ndim = 1, dtype = numpy.float64]
        The limits of the system over which the pixels span in the
        y-dimension, formatted as [ymin, ymax].

    '''
    cdef Py_ssize_t     nrows = positions.shape[0]

    cdef Py_ssize_t     nx = pixels.shape[0]
    cdef Py_ssize_t     ny = pixels.shape[1]

    cdef double         xsize = (xlim[1] - xlim[0]) / nx
    cdef double         ysize = (ylim[1] - ylim[0]) / ny

    # print((
    #     f"nrows = {nrows}\n"
    #     f"nx = {nx}\n"
    #     f"ny = {ny}\n"
    #     f"xsize = {xsize}\n"
    #     f"ysize = {ysize}\n"
    # ))

    cdef double         x, y, r         # Particle x, y, radius
    cdef double         px, py          # Pixel centre x, y

    cdef Py_ssize_t     min_nx, max_nx
    cdef Py_ssize_t     min_ny, max_ny

    cdef Py_ssize_t     num_corners     # Number of intersected corners
    cdef Py_ssize_t[4]  corners         # Intersected corners indices

    cdef Py_ssize_t     i, j, k, l

    for i in range(nrows):
        # Aliases
        x = positions[i, 0]
        y = positions[i, 1]
        r = radii[i]

        if r == 0:
            if x == xlim[1]:
                min_nx = nx - 1
            else:
                min_nx = <Py_ssize_t>floor((x - xlim[0]) / xsize)

            if y == ylim[1]:
                min_ny = ny - 1
            else:
                min_ny = <Py_ssize_t>floor((y - ylim[0]) / ysize)

            if min_nx >= nx or min_nx < 0:
                continue
            if min_ny >= ny or min_ny < 0:
                continue

            pixels[min_nx, min_ny] += 1
            continue

        # Find minimum and maximum pixel indices over which the particle spans.
        # This corresponds to the intersected cells
        min_nx = <Py_ssize_t>floor((x - r - xlim[0]) / xsize)
        max_nx = <Py_ssize_t>floor((x + r - xlim[0]) / xsize)

        min_ny = <Py_ssize_t>floor((y - r - ylim[0]) / ysize)
        max_ny = <Py_ssize_t>floor((y + r - ylim[0]) / ysize)

        # print((
        #     "Before thresholding\n"
        #     f"x = {x}\n"
        #     f"y = {y}\n"
        #     f"r = {r}\n"
        #     f"min_nx = {min_nx}\n"
        #     f"max_nx = {max_nx}\n"
        #     f"min_ny = {min_ny}\n"
        #     f"max_ny = {max_ny}\n"
        # ))

        # Check pixel indices are within the system bounds
        if min_nx >= nx:
            continue
        elif min_nx < 0:
            min_nx = 0

        if max_nx < 0:
            continue
        elif max_nx >= nx:
            max_nx = nx - 1

        if min_ny >= ny:
            continue
        elif min_ny < 0:
            min_ny = 0

        if max_ny < 0:
            continue
        elif max_ny >= ny:
            max_ny = ny - 1

        # print((
        #     "After thresholding\n"
        #     f"x = {x}\n"
        #     f"y = {y}\n"
        #     f"r = {r}\n"
        #     f"min_nx = {min_nx}\n"
        #     f"max_nx = {max_nx}\n"
        #     f"min_ny = {min_ny}\n"
        #     f"max_ny = {max_ny}\n"
        # ))


        # Iterate over all intersected pixels, computing the area covered by
        # the particle
        for j in range(min_nx, max_nx + 1):
            for k in range(min_ny, max_ny + 1):
                # Find the number of pixel corners inside the particle
                num_corners = 0
                for l in range(4):
                    corners[l] = 0

                # Pixel centre coordinates
                px = xlim[0] + (j + 0.5) * xsize
                py = ylim[0] + (k + 0.5) * ysize

                # print((
                #     f"For pixel indices [{j}, {k}]:\n"
                #     f"px = {px}\n"
                #     f"py = {py}\n"
                #     f"r2 = {r * r}\n"
                #     f"dist2 = {dist2(px - 0.5 * xsize, py - 0.5 * ysize, x, y)}\n"
                #     f"dist2 = {dist2(px + 0.5 * xsize, py - 0.5 * ysize, x, y)}\n"
                #     f"dist2 = {dist2(px + 0.5 * xsize, py + 0.5 * ysize, x, y)}\n"
                #     f"dist2 = {dist2(px - 0.5 * xsize, py + 0.5 * ysize, x, y)}\n"
                # ))

                # Lower left corner
                if dist2(px - 0.5 * xsize, py - 0.5 * ysize, x, y) < r * r:
                    num_corners += 1
                    corners[0] = 1

                # Lower right corner
                if dist2(px + 0.5 * xsize, py - 0.5 * ysize, x, y) < r * r:
                    num_corners += 1
                    corners[1] = 1

                # Upper right corner
                if dist2(px + 0.5 * xsize, py + 0.5 * ysize, x, y) < r * r:
                    num_corners += 1
                    corners[2] = 1

                # Upper left corner
                if dist2(px - 0.5 * xsize, py + 0.5 * ysize, x, y) < r * r:
                    num_corners += 1
                    corners[3] = 1

                # print((
                #     f"num_corners = {num_corners}\n"
                #     f"corners = [{corners[0], corners[1], corners[2], corners[3]}]\n"
                # ))

                if num_corners == 0:
                    pixels[j, k] += zero_corner(corners, px, py, xsize, ysize, x, y, r)
                elif num_corners == 1:
                    pixels[j, k] += one_corner(corners, px, py, xsize, ysize, x, y, r)
                elif num_corners == 2:
                    pixels[j, k] += two_corner(corners, px, py, xsize, ysize, x, y, r)
                elif num_corners == 3:
                    pixels[j, k] += three_corner(corners, px, py, xsize, ysize, x, y, r)
                elif num_corners == 4:
                    pixels[j, k] += four_corner(corners, px, py, xsize, ysize, x, y, r)
